#!/usr/bin/perl -w 

# This script is called from the DRMS library when there is a request for an SU that
# is not online and belongs to a remote site. It calls a cgi program jsoc_fetch
# with op=exp_su to request, from the remote site, an SU. If the SU is online,
# jsoc_fetch will get the required information (sunums, series, and file paths)
# from a SUM_info() call, and then generate a response file (a .txt file specified
# by $kGETAPPOUT below). If the SU is offline, jsoc_fetch will then submit
# a request to jsoc_export_manage (by adding a record to jsoc.export_new). Either
# way, jsoc_fetch will generate the response file. remotesums_master.pl
# will then parse the response file to examine the status. If status == 0,
# jsoc_fetch handled the request synchronously and in the response file there
# will exist the desired sunums, series, and file paths. remotesums_master.pl will
# then hand this information to remotesums_ingest to scp the files locally. If status != 0,
# then the request is being handled asynchronously (by jsoc_export_manage). In that
# case, remotesums_master.pl will then poll by sending a new jsoc_fetch request
# with op=exp_status. This request will also result in a response file. remotesums_master.pl
# will keep polling by sending these jsoc_fetch requests until status == 0 in the
# response. When that happens, sunums, series, and file paths will be available
# in the response file, and again, those are sent to remotesums_ingest for scp
# processing.



# The cmd takes the following form:
#   remotesums_master.pl <url> '=' <sunum> ',' <sunum...> '#' <url> '=' <sunum> ',' <sunum...> # ...
#   Example:
#     remotesums_master.pl http://jsoc.stanford.edu/cgi-bin/ajax/jsoc_fetch=series1{1234567,1234568}&series2{1234569}#http://jsoc.stanford.edu/cgi-bin/ajax/jsoc_fetch2=series1{7654321,7654320}

# When calling this script from DRMS, print error messages to STDERR - STDOUT is redirected
# to a pipe back to a parent process in DRMS.

# ssh-agent must be run, and the check for an instance of ssh-agent, must be done by
# the production user that starts sum_export_svc &.
# The proper way to set up ssh-agent for use to obtain files from Stanford:
#   First, create passphrased private (id_rsa)/public keys (id_rsa.pub) and place the 
#   public key content in j0.stanford.edu's authorized_keys file. Then:
#     ssh-agent -c > ~/.ssh-agent_rs
#     source ~/.ssh-agent_rs
#     ssh-add ~/.ssh/id_rsa (enter passphrase)
#

use CGI;
use Data::Dumper;
use LWP::UserAgent;
use HTTP::Request::Common;




# Global defines
#open LOG, ">>/tmp/remotesums.log";

#print LOG $$,"::",`date +'%Y%m%d.%H%M%S'`,"\n";
#print LOG Dumper [@ARGV];

# Get arguments
if (scalar(@ARGV) != 1)
{
    #print STDERR "A single argument, a list of SUNUMs, is expected.\n";
#    print LOG "A single argument, a list of SUNUMs, is expected.\n";
    print STDERR "A single argument, a list of SUNUMs, is expected.\n";
#    close LOG;
    exit(1);
}

#close LOG;

$arg = $ARGV[0];


## break down sequence
my @urls = 

my @series_sunum_seq=();
map {
       if ($_ =~ /(\S+)=(\S+)/ ) {
          push @series_sunum_seq, $2;
       } 
    } split ("#", $arg);


##  Breaking the urls into series/sunum chunks
my @series_sunum_chunk = ();
map { 
      push @series_sunum_chunk, split('&',$_);
    } @series_sunum_seq;


## request hash tree
## The keyword is the "series" name and its value an array of sunums
my $req_tree = {};

map {
      if ( $_ =~ /(\S+)\{([\d,]+)\}/ ) {
        $req_tree->{$1}=[] unless exists $req_tree->{$1};
        push @{$req_tree->{$1}}, split(",",$2);
      }
    } @series_sunum_chunk;


#print Dumper $req_tree;

## request to the JMD
my $priority=100;
for my $series (keys %{$req_tree}) {
  $sreq={"type"=> "request", "series"=> [$series], "$series" => $req_tree->{$series}, priority=>$priority}; 
#  print Dumper $sreq;
  my $ret = get_sunums ($sreq);
  ## if the submission was successfull
  ## wait till the sunum has been downloaded.
  my $total_wait = 0;
  my $RETRIES=20;
#  if ($ret == 0) {
    my $waiting = 1;
    while ($waiting) {
      my ($waitNo, $fails) = get_sunum_status($req_tree->{$series});
      if ($waitNo > 0 && $fails == 0 && $total_wait < $RETRIES) {
        $total_wait ++;
        sleep 5;
      } elsif ( $fails > 0 || $total_wait >=$RETRIES ) {
        $waiting = 0;
        print "0\n";
      } else {
        $waiting = 0;
        print "0\n";
      }
    }
#  } else {
#    my ($waitNo, $fails) = get_sunum_status($req_tree->{$series});
#    if ($fails ==0 && $waitNo == 0 ) {
#      print "1\n";
#    } else {
#      print "0\n";
#    }
#  }
}

sub jmd_call {
  my $sreq = shift;
  my $ua = LWP::UserAgent->new;
  $ua->agent("SOAP::Lite/Perl/0.55");
  my $url='http://<JMD machine name here>:8080/JMD/JMD';

  my $res = $ua->post($url,$sreq);

  return $res;
}

sub get_sunums {
  my $sreq = shift;

  my $res = jmd_call($sreq);

#  open LOG, ">>/tmp/remotesums.log";
#  print LOG "get_sunums::content [$sreq]:\n", $res->content;
#  print LOG "get_sunums::content [", Dumper($sreq), "]\n";
#  close LOG;
  if ( $res->content =~ /SUCCESFUL submissions:(\d+)/) {
    my $sucessful=$1;
    my $submissions=scalar @{$sreq->{$sreq->{"series"}->[0]}};
    return ($sucessful != $submissions)? 1: 0; 
  } else {
    unless ($res->is_success) {
      print STDERR "ERROR::\n";
      print STDERR $res->status_line, "\n";
    }
    return 1;
  }

}

sub get_sunum_status {

  my $sunums = shift;
  my $sreq={"type"=> "query", "sunums" => $sunums};

  my $res = jmd_call($sreq);

#  open LOG, ">>/tmp/remotesums.log";
#  print LOG "content [", @{$sunums}, "] status:\n", $res->content;
  #print LOG "content [", Dumper($sunums), "] status:\n", $res->content;
#  close LOG;
#sunum=26846541;state=DONE
#sunum=29966673;state=DONE
  my $done = 0;
  my $fail = 0;
  for my $sunum (@{$sunums}) {
    if ($res->content =~ /sunum=$sunum;state=(\w+)/) {
      my $status = $1;
      if ($status =~ /DONE/ ) {
        $done++;
      } elsif ( $status =~/FAIL|TOUT|N\/A|EROR|INVD|OFLN/) {
        $fail++;
      }
     
    }
  }

  return (scalar (@{$sunums}) - $done, $fail);
}

