#include "jsoc_main.h"
#include "drms_types.h"
#include "drms_storageunit.h"
#include "exputil.h"
#include "fitsexport.h"

//ISS fly-tar START
#include "fitsio.h"
#include "cfitsio.h"
#include "libtar.h"
#define  MAXTARSIZE 2147483648
//#define  MAXTARSIZE 9910560
#ifndef __export_callback_func_t__
#define __export_callback_func_t__
typedef int (*export_callback_func_t)(char *, ...); //ISS fly-tar
#endif
int flytar_append_file_header(TAR *t, char * filename, int size);
int flytar_append_file_tail(TAR *t, int size);
int flytar_append_file(TAR *t, char * fullname);
int file_set_stats(TAR *t, int size, char * filename);
int block_rest(int block_size, int file_size);
char * replace_blanks(char *str1, const char *str2, int n);
int match_query_segments(DRMS_Segment_t *seg);
char ** parse_query_for_segment_names(char * query);
int count_segment_names(char **segments);
int testRecordCount(DRMS_RecordSet_t* rsin, int* stat, int querySegCount);
void CallbackErrorBuffer( char* filename, char*query, int errorCode);
void CallbackClose(const char * rsquery, TAR **tar, int recCount);
void CallbackInfo( char* filename);

char sackfile[4000];
char starfile[4000];
char **segments       =    (char **) NULL; //work-around
//ISS fly-tar END

/**
@defgroup expfits jsoc_export_as_fits - Export internally stored data to a FITS file 
@ingroup su_export

@brief This module exports DRMS keywords and data segments into one or more FITS files. The resulting files are completely self-contained. DRMS keywords have been converted to FITS keywords. The caller can specify the compression parameters to use when creating the FITS file. 

@par Synopsis:
@code
stand-alone mode:
jsoc_export_as_fits rsquery=<recset query> n=<limit> reqid=<export request id> expversion=<version> 
     method=<exp method> protocol=<output-file protocol> path=<output path&gt 
     { ffmt=<filename format> } { kmclass=<keymap class> } { kmfile=<keymap file> } 
     { cparms=<compression string list> }

or

export-series mode:
jsoc_export_as_fits reqid=<export request id> expversion=<version> method=<exp method> 
     protocol=<output-file protocol> { expseries=<exp series> } { kmclass=<keymap class> } 
     { kmfile=<keymap file> } { cparms=<compression string list> }
@endcode

For each segment for each record specified by a record-set query, this module creates a FITS file. The FITS 
file contains a copy of the image data that was contained in the original internal SUMS file. It also 
contains FITS keywords derived via conversion from the original DRMS keywords. The conversion process
is described in detail in ::drms_keyword_getmappedextname. Succinctly, if @a kmfile is provided, 
the argument identifies a file which contains a keyword map which is used to map the DRMS keywords
to FITS keywords. If @a kmfile is not provided, then if @a kmclass is provided, 
this argument identifies a keyword-map class, which is a name that identifies a predefined 
keyword map which is then used to map the DRMS keywords to FITS keywords. Otherwise, if the 
DRMS keyword being converted has a description field that contains
a string of the form [X{:Y}], then if X is a valid FITS keyword, it is used. Otherwise, if the 
DRMS keyword name is a valid FITS keyword name, it is used. Otherwise, an algorithm is used to automatically
generate a FITS keyword name.

The name of the output file is determined from a filename format string. The format contains a series
of substrings, each of which starts with '{', is followed by either "seriesname", "segment", "recnum",
or "recnum:<format specifier>;", and ends with '}'. Each of these strings may appear
as many times and in any order in the overall format string.
Each of these strings serves as a placeholder that will be replaced by 
an appropriate string value. For example, if the file being exported originates from a series named
"su_arta.testfits", then "{seriesname}" will be evaluated to "su_arta.testfits". "{segment}"
evaluates to the name of the segment being exported, and "{recnum}" evaluates to the 
record number of the record being exported. The optional &lt;format specifier&gt; in 
"{recnum:<format specifier>}" is the printf format specifier used to convert the record number into 
a string (by default "%lld" is used if the format specifier is ommitted). As an example, to
generate a filename like mdi.fd_M_96m_lev18.1111.data.fits, the format string 
"{seriesname}.{recnum:%d}.{segment}" would suffice.

This module provides two ways to export data. When run in "stand-alone" mode, the exported FITS
files are written to a directory specified by the @a path argument, or to the working directory 
if no @a path argument is present. To use this mode, the @a rsquery argument must be present. This
query identifies the DRMS records and segments to be exported. Under this mode, the module does not assume 
the existence of an export series (see "export-series" mode). It simply exports to a 
specified directory. As such, the
@a expseries argument is ignored if present. An optional @a ffmt argument contains the 
filename format string. If this optional argument is not present, a default
format string of "{seriesname}.{recnum:%lld}.{segment}" is assumed.

When the module is run in "export-series" mode, 
the record-set query that identifies records to export 
originates from an "export series", not from the command-line. An export series is a 
DRMS series established by an administrator. It is used to track exports, and in fact the 
SUNUM values of this series' records point to directories where the exported FITS files 
temporarily reside. A dbase query for the
request ID (every export request is assigned a unique request ID) in the export series 
returns a record which contains the
record-set query and the output filename format string (at the very least, the keywords 
rsquery - the record-set query - and ffmt - the filename format string - 
must exist in the export series). This export-series record is cloned, and the 
output files are written to the clone's SUDIR. 
To use the export-series mode, the @a rsquery argument must NOT be present. The export series
is identified by the @a expseries argument if it is present, otherwise it defaults to jsoc.export.
Please see ???? for more information about the general export process. 

@a cparms is a comma-separated list of strings. Each string is either 
a CFITSIO compression string (please see tile-compression arguments in 
http://jsoc.stanford.edu/jsocwiki/Jsd for more information 
about this string), or the string "**NONE**". "**NONE**" indicates that the
output file(s) should not be compressed. Each string in this list is relevant to 
one of the segments being exported, the order of the strings matches the order of the 
segments as determined by the original .jsd used to create the segment.

This module writes, into the export directory, a "packing list" 
which is a file that contains metadata about the export. This
information includes the request ID of the export, the version of the export code, 
the export method, and the export protocol. These data are passed into the module via the
@a reqid, @a expversion, @a method, and @a protocol arguments, which are required regardless
of the mode under which the module is run. The packing list also contains five other 
metadata which are caculated/determined during the module run: a count of the total 
number of files exports, the total size in bytes of the exported files, the time of
export completion, the directory to which the files were exported (provided by the @a path
argument to the module), and the status of the export (0 implies success). Finally, 
the packing list contains a two-column table of output files. The first column
is a record-set query, and the second column is a single exported file. The 
record-set query uniquely identifies the output file.

@par Flags:
This module has no flags.

@par GEN_FLAGS:
Ubiquitous flags present in every module.
@ref jsoc_main

@param rsquery A DRMS record-set query that identifies a set of records and segments
to export.
@param reqid A string that uniquely identifies an export request. This string appears
in the output packing list.
@param expversion The version of the export code calling this module. This string appears
in the output packing list.
@param method The method of export, such as url or url_quick. This string appears
in the output packing list.
@param n Record count limit.
@param protocol The type of output file to produce. This must be FITS. This string appears
in the output packing list.
@param path The directory to which output files are to be written. If present, this string appears
in the output packing list.
@param ffmt The filename format string which defines the name of the output file(s).
@param kmclass Identifies a keyword-map class, which is a name that identifies a predefined 
keyword map which is then used to map the DRMS keywords to FITS keywords.
@param kmfile  Identifies a file which contains a keyword map which is used to map the DRMS keywords
to FITS keywords. 
@param cparms A comma-separated list of strings. Each string is either 
a CFITSIO compression string or the string "**NONE**".

@par Exit_Status:
@c 0 success<br>
@c 1 cannot register definition file<br>
@c 2 missing required argument<br>
@c 3 invalid request ID<br>
@c 4 invalid record-set query<br>
@c 5 invalid filname format string<br>
@c 6 failure to write export file<br>
@c 7 cannot open packing list file<br>
@c 8 failure writing packing list<br>
@c 9 unsupported packing list element type

@par Example:
Running in the stand-alone mode:
@code
jsoc_export_as_fits reqid=JSOC_20090514_001 expversion=0.5 rsquery='su_phil.fd_M_daily[1996.07.4]' 
     path=/tmp/jsocexptest ffmt='{seriesname}.{recnum:%d}.{segment}' method=url_quick 
     protocol=FITS cparms="**NONE**"
@endcode
*/

char *module_name = "export";

typedef enum
{
   kMymodErr_Success = 0,
   kMymodErr_CantRegisterDefs,
   kMymodErr_MissingArg,
   kMymodErr_BadRequestID, 
   kMymodErr_BadRecSetQuery,
   kMymodErr_BadFilenameFmt,
   kMymodErr_ExportFailed,
   kMymodErr_CantOpenPackfile,
   kMymodErr_PackfileFailure,
   kMymodErr_UnsupportedPLRecType,
   kMymodErr_DRMS,
   kMymodErr_MissingSegFile
} MymodError_t;

typedef enum
{
   kPL_metadata,
   kPL_content
} PLRecType_t;

#define kNotSpecified    "NOT SPECIFIED"
#define kArg_reqid       "reqid"
#define kArg_version     "expversion"
#define kArg_method      "method"
//ISS fly-tar START
#define kArg_ackfile     "ackfile"
#define kArg_tarfile     "tarfile"
#define kArg_segments    "segments"
//ISS fly-tar END
#define kArg_protocol    "protocol"
#define kArg_rsquery     "rsquery"
#define kArg_n      	 "n"
#define kArg_expSeries   "expseries"
#define kArg_ffmt        "ffmt"
#define kArg_path        "path"
#define kArg_clname      "kmclass"
#define kArg_kmfile      "kmfile"
#define kArg_cparms      "cparms"


#define kDef_expSeries   "jsoc.export"

#define kPWD             "PWD"

#define kNoCompression   "**NONE**"

#define kMB              (1048576)

/* If rsquery is provided as a cmd-line argument, then jsoc_export does not 
 * save the output data files to an export series.  Instead the caller
 * MUST provide a filename format string (which is a template that 
 * describes how to create the output file names).  The caller may also
 * provide the path argument to specify the directory into which the 
 * output files are to be written.  If the path argument is not specified,
 * then the output files are written to the current working directory.
 * If requery is omitted from the cmd-line, then an export series
 * must either be provided on the cmd-line, or the default series is used.
 * The query that determines what files are to be exported is found
 * in the export series by searching for the record with a matching
 * reqid.
 */
ModuleArgs_t module_args[] =
{
     {ARG_STRING, kArg_version, "", "jsoc export version."},
     {ARG_STRING, kArg_reqid, "", 
        "Export series primary key value that identifies the output record."},
     {ARG_STRING, kArg_method, "", "jsoc export method (eg, url or ftp)."},
     //ISS fly-tar START
     {ARG_STRING, kArg_ackfile, kNotSpecified, "Acknowledgement File"},
     {ARG_STRING, kArg_tarfile, kNotSpecified, "Tar File Name"},
     {ARG_STRING, kArg_segments, kNotSpecified, "Export segment name."},
     //ISS fly-tar END
     {ARG_STRING, kArg_protocol, "", "file conversion method (eg, convert to fits)."},
     {ARG_STRING, kArg_rsquery, kNotSpecified, 
        "Record-set query that specifies data to be exported."},
     {ARG_STRING, kArg_expSeries, kDef_expSeries, "Series to which exported data are saved."},
     {ARG_STRING, kArg_ffmt, kNotSpecified, "Export filename template."},
     {ARG_STRING, kArg_path, kNotSpecified, "Path to which fits files are output."},
     {ARG_STRING, kArg_clname, kNotSpecified, "Export key map class."},
     {ARG_STRING, kArg_kmfile, kNotSpecified, "Export key map file."},
     {ARG_STRING, kArg_cparms, kNotSpecified, "FITS-stanford compression string used to compress exported image."},
     {ARG_INT, kArg_n, "0", "Record count limit."},
     {ARG_END}
};

char gDefBuf[PATH_MAX] = {0};

/* Convert number of bytes to number of MB. If number of bytes < 1MB, then return 1. */
static long long ToMB(long long nbytes)
{
   if (nbytes <= kMB)
   {
      return 1;
   }

   return nbytes / kMB;
}

MymodError_t WritePListRecord(PLRecType_t rectype, FILE *pkfile, const char *f1, const char *f2)
{
   MymodError_t err = kMymodErr_Success;

   switch (rectype)
   {
      case kPL_metadata:
        fprintf(pkfile, "%s=%s\n", f1, f2);
        break;
      case kPL_content:
        fprintf(pkfile, "%s\t%s\n", f1, f2);
        break;
      default:
        fprintf(stderr, "Unsupported packing-list record type '%d'.\n", (int)rectype);
        err = kMymodErr_UnsupportedPLRecType;
   }

   return err;
}

/* Assumes tcount is zero on the first call.  This function adds 
 * the number of files exported to tcount on each call. */
static int MapexportRecordToDir(DRMS_Record_t *recin,
                                const char *ffmt,
                                const char *outpath,
                                FILE *pklist,
                                const char *classname, 
                                const char *mapfile,
                                int *tcount, 
                                const char **cparms,
                                MymodError_t *status,
                                export_callback_func_t callback) //ISS fly-tar
{
   int drmsstat = DRMS_SUCCESS;
   MymodError_t modstat = kMymodErr_Success;
   DRMS_Segment_t *segin = NULL;
   unsigned long long tsize = 0;
   char dir[DRMS_MAXPATHLEN];
   char fmtname[DRMS_MAXPATHLEN];
   char fullfname[DRMS_MAXPATHLEN];
   char query[DRMS_MAXQUERYLEN];
   struct stat filestat;
   HIterator_t *last = NULL;
   int iseg;
   int lastcparms;
   int gotone;

   drms_record_directory(recin, dir, 1); /* This fetches the input data from SUMS. */

   /* Must create query from series name and prime keyword values */
   drms_sprint_rec_query(query, recin);

   /* The input rs query can specify a subset of all the series' segments - 
    * this is encapsulated in recin. */

   iseg = 0;
   lastcparms = 0;
   gotone = 0;

   while ((segin = drms_record_nextseg(recin, &last, 1)) != NULL)
   {
      if (exputl_mk_expfilename(segin, ffmt, fmtname) == kExpUtlStat_Success)
      {
         snprintf(fullfname, sizeof(fullfname), "%s/%s", outpath, fmtname);
      }
      else
      {
         modstat = kMymodErr_BadFilenameFmt;
         break;
      }

      if (!cparms || !cparms[iseg])
      {
         lastcparms = 1;
      }

      drmsstat = fitsexport_mapexport_tofile2( segin, 
                                               !lastcparms ? cparms[iseg] : NULL, 
                                               classname, 
                                               mapfile, 
                                               fullfname,
                                               callback); //ISS fly-tar

      //ISS fly-tar : modify test on file only if callback NULL
      if ( drmsstat != DRMS_SUCCESS && callback != NULL) {  //ISS fly-tar
         fprintf(stderr, "(1) Failure exporting segment '%s'.\n", segin->info->name);
         (*callback)("error",query,fullfname);  //ISS fly-tar
      }
      else if (drmsstat == DRMS_ERROR_INVALIDFILE)
      {
         /* No input segment file. */
      }
      else if ((drmsstat != DRMS_SUCCESS ||
               ( stat(fullfname, &filestat) ) && (callback == NULL)) ) //ISS fly-tar
      {
         /* There was an input segment file, but for some reason the export failed. */
         modstat = kMymodErr_ExportFailed;
         fprintf(stderr, "(2) Failure exporting segment '%s'.\n", segin->info->name);
         break;
      }
      else
      {
         //ISS fly-tar
         if (callback != NULL) {
           (*callback)("info",fullfname);
         } else {

           gotone = 1;

           if (tcount)
           {
              ++*tcount;
           }

           tsize += filestat.st_size;
           WritePListRecord(kPL_content, pklist, query, fmtname);
         } //ISS fly-tar
      }

      iseg++;
   }

   /* If NO file exported, this is an error. But if there is more than one segments, it is okay
    * for some input segment files to be missing. */
   if (!gotone && callback == NULL)
   {
      modstat = kMymodErr_ExportFailed;
   }

   if (last)
   {
      hiter_destroy(&last);
   }

   if (status)
   {
      *status = modstat;
   }

   return tsize;
}


//ISS fly-tar
char    *infobuffer     =    (char *) NULL;
char    *errorbuffer    =    (char *) NULL;
int      terminate      = 0;
int      print_header   = 0;

static int MapexportToDir(DRMS_Env_t *env, 
                          const char *rsinquery, 
                          const char *ffmt,
                          const char *outpath, 
                          const char *method,    //ISS fly-tar
                          FILE *pklist, 
                          const char *classname, 
                          const char *mapfile,
                          int *tcount,
                          TIME *exptime,
                          const char **cparms, 
                          MymodError_t *status)
{
   int stat = DRMS_SUCCESS;
   MymodError_t modstat = kMymodErr_Success;
   DRMS_RecordSet_t *rsin = NULL;
   DRMS_Record_t *recin = NULL;
   int iRec = 0;
   int nSets = 0;
   int iSet = 0;
   int nRecs = 0;
   int RecordLimit = *tcount;
   unsigned long long tsize = 0;
   int errorCount = 0;
   int okayCount = 0;

   //ISS fly-tar START
   export_callback_func_t callback = NULL;

   int first = 1;
   int recCount = 0;
   //int testCount = 0;
   TAR *tar = NULL;
   char tarFileNameElement[255];
//   char *segments[]= (char **) NULL;
   int segmentCount=0;
   //ISS fly-tar END

   if (RecordLimit == 0)
     rsin = drms_open_records(env, rsinquery, &stat);
   else
     rsin = drms_open_nrecords(env, rsinquery, RecordLimit, &stat);

   if (rsin)
   {
      //ISS fly-tar START

      if (strcmp (method, "url_cgi") == 0) {
        char fitsfilename[3000];
        recCount=testRecordCount(rsin,&stat,segmentCount);
        if (recCount > 1) {

          int my_open (const char *path, int flags, ... ) {
             fflush(NULL);
            return 1;
          }

          tartype_t my_type = { my_open, close, read, write };

             //create tar
          if (tar_open(&tar, "some_generic_name.tar", &my_type,
                 O_WRONLY | O_CREAT, 0644, TAR_GNU) == -1 ) {
            fprintf(stderr, "tar_open(): %s\n", strerror(errno));
            return -1;
          }
          fflush(NULL);
        }

        void CGICallback(char *callbackState, ...) {
           va_list arguments;
           va_start(arguments, callbackState);
           static DRMS_Array_t *arrout = NULL;
           static long totalSize = 0;

           if (strcmp(callbackState,"create")==0) {

             fitsfile **fptr           = (fitsfile **)  va_arg(arguments, fitsfile **);
             char      *filein         = (char *)       va_arg(arguments, char *);
             char      *cparms         = (char *)       va_arg(arguments, char *);
             int       *cfiostat       = (int *)        va_arg(arguments, int  *);
             int       *retVal         = (int *)        va_arg(arguments, int  *);

             char      *tempfilein     = "mem://";
             char       filename[252];

             strcpy(fitsfilename,filein); //hack ISS TODO
 

             if (first && (1 < recCount)) {
               char tarfilename[300];
               print_header=1;
               printf("Content-type: application/octet-stream\n");
               printf("Content-Disposition: attachment; filename=\"%s.tar\"\n",strlen(starfile)==0?replace_blanks(tarfilename,rsinquery,300):starfile);
               printf("Content-transfer-encoding: binary\n\n");
               first = 0;
             }
             //fprintf(stderr, "after reading params filein=[%s]\n", filein);
             //save file element locally for second call of the callBack function when a tar file exists.
             strcpy(tarFileNameElement, filein);

             if (cparms && *cparms) {
               snprintf(filename, sizeof(filename), "%s[%s]", tempfilein, cparms);
             } else {
               snprintf(filename, sizeof(filename), "%s", tempfilein);
             }
              
             //fprintf(stderr, "before fits_create_file\n");
             *retVal=fits_create_file(fptr,filename, cfiostat);

           } else if (strcmp(callbackState,"stdout")==0) {
             fitsfile *fptr    = (fitsfile *) va_arg(arguments, fitsfile *);
             void *image       = (void *)      va_arg(arguments, void *);
             int *retVal       = (int *)      va_arg(arguments, int *);
             fitsfile *outfptr = NULL;
             int mystatus      = 0;
             int outcfiostat   = 0;
          
             if (first && (1 == recCount)) {
               char *basename = strrchr(fitsfilename, '/');
               basename = basename? basename + 1: (char *) fitsfilename;
               print_header=1;
               printf("Content-type: application/octet-stream\n");
               printf("Content-Disposition: attachment; filename=\"%s\"\n",basename);
               printf("Content-Length: %d\n",fptr->Fptr->logfilesize);
               printf("Content-transfer-encoding: binary\n\n");
               first=0;
             }

             fprintf(stderr, "\nfile [%s], file size [%d]\n",tarFileNameElement, fptr->Fptr->logfilesize);
             if (totalSize <= MAXTARSIZE) {

               if (recCount > 1) {
                 fflush(NULL);
                 totalSize += fptr->Fptr->logfilesize;
                 if (flytar_append_file_header(tar, tarFileNameElement, fptr->Fptr->logfilesize)){
                   fprintf(stderr, "Error in flytar_append_file\n");
                   *retVal = -1;
                 }
               }
  
               //free some memory
               if (arrout->data != NULL) {
                 free(arrout->data);
                 arrout->data=NULL;
                 arrout=NULL;
               }
  
  
               fflush(NULL);
               //Dump fits mem file to stdout
               int cpstatus=0;
               int hdunum=0;
               fits_get_num_hdus(fptr, &hdunum, &cpstatus);
               //fprintf(stderr,"Number of hdus is %d\n",hdunum);
               if (hdunum > 0) {
                 for (int i=1; i<=hdunum; i++) {
                   int hdutype=0;
                   fits_movabs_hdu(fptr, i, &hdutype, &cpstatus);
                   fits_write_hdu(fptr, stdout, &cpstatus);
                 }
               }
               
               fflush(NULL);
               if (recCount > 1) {
                 if (flytar_append_file_tail(tar, fptr->Fptr->logfilesize)) {
                   fprintf(stderr, "Error in flytar_append_tail\n");
                   *retVal = -1;
                 }
               }
  
               fits_close_file(fptr, &mystatus);
             } else {
               terminate =1;
               fits_close_file(fptr, &mystatus);
               CallbackErrorBuffer( tarFileNameElement, "Size limitation in tar file.User rice compression or/and make a smaller query", 2);
             }
             fprintf(stderr,"terminate --stdout is [%d]\n",terminate);
           } else if (strcmp(callbackState,"info")==0) {
             char      *filename         = (char *)       va_arg(arguments, char *);
             fprintf(stderr,"terminate --info is [%d]\n",terminate);
             CallbackInfo(filename);

           } else if (strcmp(callbackState,"error")==0) {
             char      *query            = (char *)       va_arg(arguments, char *);
             char      *filename         = (char *)       va_arg(arguments, char *);
             fprintf(stderr,"terminate -- error is [%d]\n",terminate);
             CallbackErrorBuffer( filename, query, 0);
           } else if (strcmp(callbackState,"close")==0) {
             fprintf(stderr,"terminate --close is [%d]\n",terminate);
             CallbackClose(rsinquery, &tar, recCount);
           } else if (strcmp(callbackState,"setarrout")==0) {
             arrout    = (DRMS_Array_t *) va_arg(arguments, DRMS_Array_t *);
           }
         }
         callback = (export_callback_func_t) CGICallback;
         fprintf(stderr,"Multiple Record Count [%d]\n",recCount);
      }
      //ISS fly-tar END

      /* stage records to reduce number of calls to SUMS. */
      drms_stage_records(rsin, 1, 0);
      nSets = rsin->ss_n;

      for (iSet = 0; iSet < nSets; iSet++)
      {
         nRecs = drms_recordset_getssnrecs(rsin, iSet, &stat);


         if (stat != DRMS_SUCCESS)
         {
            fprintf(stderr, "Failure calling drms_recordset_getssnrecs(), skipping subset '%d'.\n", iSet);
         }
         else
         {
             for (iRec = 0; iRec < nRecs; iRec++)
             {
                recin = rsin->records[(rsin->ss_starts)[iSet] + iRec];
                tsize += MapexportRecordToDir(recin,
                                              ffmt,
                                              outpath,
                                              pklist, 
                                              classname, 
                                              mapfile, 
                                              tcount, 
                                              cparms, 
                                              &modstat,
                                              callback);  //ISS fly-tar

                if (modstat == kMymodErr_Success)
                {
                   okayCount++;
                }
                else
                {
                   errorCount++;
                }

                //ISS fly-tar
                if (terminate) {
                   (*callback)("close");
                   break;
                }
             }

            //ISS fly-tar
             if (terminate)
               break;
         }
      }
	 
      modstat = kMymodErr_Success; /* Could have been set to BAD in loop above, but okayCount and errorCount
                                    * account for fatal errors in the loop, not modstat. */

      if (errorCount > 0)
      {
         fprintf(stderr,"Export Failed for %d segments of %d attempted.\n", errorCount, errorCount + okayCount);
         if (okayCount == 0)
           modstat = kMymodErr_ExportFailed;
      }

      if (exptime)
      {
         *exptime = CURRENT_SYSTEM_TIME;
        //ISS fly-tar START
        if (strcmp (method, "url_cgi") == 0) {
          fprintf(stderr, "calling close ... \n");
          (*callback)("close");
        }
        //ISS fly-tar END
      }
   }
   else
   {
      fprintf(stderr, "Record-set query '%s' is not valid.\n", rsinquery);
      modstat = kMymodErr_BadRecSetQuery; 
   }

   if (rsin)
   {
      drms_close_records(rsin, DRMS_FREE_RECORD);
   }

   if (status)
   {
      *status = modstat;
   }
   
   return tsize;
}

static MymodError_t CallExportToFile(DRMS_Segment_t *segout, 
                                     DRMS_Segment_t *segin, 
                                     const char *clname,
                                     const char *mapfile,
                                     const char *ffmt,
                                     unsigned long long *szout,
                                     char *filewritten,
                                     const char *cparms)
{
   int status = DRMS_SUCCESS;
   MymodError_t err = kMymodErr_Success;
   char fileout[DRMS_MAXPATHLEN];
   char filein[DRMS_MAXPATHLEN];
   char basename[DRMS_MAXPATHLEN];
   unsigned long long size = 0;
   struct stat filestat;

   if (segout)
   {
      drms_segment_filename(segin, filein); /* input seg file name */
      if (!stat(filein, &filestat))
      {
	 size = filestat.st_size;

         if (exputl_mk_expfilename(segin, ffmt, basename) == kExpUtlStat_Success)
         {
            CHECKSNPRINTF(snprintf(segout->filename, DRMS_MAXSEGFILENAME, "%s", basename), DRMS_MAXSEGFILENAME);
            drms_segment_filename(segout, fileout);

            status = fitsexport_mapexport_tofile2(segin, cparms, clname, mapfile, fileout, (export_callback_func_t) NULL);
            if (status == DRMS_ERROR_INVALIDFILE)
            {
               /* No input file for segment - not necessarily an error. */
               err = kMymodErr_MissingSegFile;
            }
            else if (status != DRMS_SUCCESS)
            {
               err = kMymodErr_ExportFailed;
               fprintf(stderr, "Failed to export segment '%s' to '%s'.\n", segin->info->name, fileout);
            }
         }
         else
         {
            err = kMymodErr_BadFilenameFmt;
         }
      }
      else
      {
	 fprintf(stderr, "Unable to open source file '%s'.\n", filein);
         err = kMymodErr_ExportFailed;
      }
   }

   *szout = 0;
   if (err == kMymodErr_Success)
   {
      *szout = size;
      snprintf(filewritten, DRMS_MAXPATHLEN, "%s", basename);
   }

   return err;
}

/* recout is the export series' record to which the exported data will be saved. */
static int MapexportRecord(DRMS_Record_t *recout,
                           DRMS_Record_t *recin,
                           const char *classname, 
                           const char *mapfile,
                           const char *pklfilename, 
                           int *tcount,
                           const char *ffmt, 
                           char **outpath,
                           FILE **pklist,
                           const char **cparms, 
                           MymodError_t *status)
{
   MymodError_t err = kMymodErr_Success;
   HIterator_t *last = NULL;
   DRMS_Segment_t *segout = NULL;
   DRMS_Segment_t *segin = NULL;
   unsigned long long size = 0;
   unsigned long long tsize = 0;
   char dir[DRMS_MAXPATHLEN];
   int iseg;
   int lastcparms;
   int gotone;

   segout = drms_segment_lookupnum(recout, 0);

   drms_record_directory(recin, dir, 1); /* This fetches the input data from SUMS. */

   if (segout)
   {
      char glom[128];
      snprintf(glom, sizeof(glom), "%%s/%s", DRMS_SLOTDIR_FORMAT);
      char path[DRMS_MAXPATHLEN];
      snprintf(path, sizeof(path), glom, segout->record->su->sudir, segout->record->slotnum);

      if (outpath)
      {
         *outpath = strdup(path);
      }

      if (pklist)
      {
         char pkpath[DRMS_MAXPATHLEN];
         snprintf(pkpath, sizeof(pkpath), "%s/%s", path, pklfilename);
         *pklist = fopen(pkpath, "w+");
      }

      char fname[DRMS_MAXPATHLEN];
      char query[DRMS_MAXQUERYLEN];

      /* Must create query from series name and prime keyword values */
      char buf[1028];
      int nkeys = recin->seriesinfo->dbidx_num;
      int snext = 0;
      int ikey = 0;

      snprintf(query, sizeof(query), "%s", recin->seriesinfo->seriesname);
      snext = strlen(query);

      for (ikey = 0; ikey < nkeys; ikey++)
      {
         drms_keyword_snprintfval(recin->seriesinfo->dbidx_keywords[ikey], buf, sizeof(buf));
         snprintf(query + snext, sizeof(query) - snext, "[%s]", buf);
         snext += strlen(buf) + 2;
      }

      /* The input rs query can specify a subset of all the series' segemnts - 
       * this is encapsulated in recin. */
      iseg = 0;
      lastcparms = 0;
      gotone = 0;

      while ((segin = drms_record_nextseg(recin, &last, 1)) != NULL)
      {
	 size = 0;

         if (!cparms || !cparms[iseg])
         {
            lastcparms = 1;
         }

	 err = CallExportToFile(segout, 
                                segin, 
                                classname, 
                                mapfile, 
                                ffmt, 
                                &size, 
                                fname, 
                                !lastcparms ? cparms[iseg] : NULL);



         if (err == kMymodErr_MissingSegFile)
         {
            /* No input file for this segment */
         }
         else if (err != kMymodErr_Success)
	 {
	    fprintf(stderr, "(3) Failure exporting segment '%s'.\n", segin->info->name);
	    break;
	 }
	 else
	 {
            gotone = 1;

            if (tcount)
            {
               ++*tcount;
            }

            tsize += size;
            if (pklist && *pklist)
            {
               WritePListRecord(kPL_content, *pklist, query, fname);
            }
	 }

         iseg++;
      }

      /* If NO file exported, this is an error. But if there is more than one segments, it is okay
       * for some input segment files to be missing. */
      if (!gotone)
      {
         err = kMymodErr_ExportFailed;
         fprintf(stderr, "Failure exporting record number %lld.\n", recin->recnum);
      }

      if (last)
      {
         hiter_destroy(&last);
      }
   }
   else
   {
      fprintf(stderr, "Export series contains no segment!\n");
      err = kMymodErr_ExportFailed;
   }

   if (status)
   {
      *status = err;
   }

   return tsize;
}		

/* recout is the export series' record to which the exported data will be saved. 
 * reqid is the primary key in the export series 
 */
static int Mapexport(DRMS_Env_t *env,
                     const char *reqid,
                     const char *classname, 
                     const char *mapfile,
                     const char *expseries,
                     const char *pklfilename, 
                     int *tcount, 
                     char **outpath, 
                     TIME *exptime, 
                     FILE **pklist, 
                     const char **cparms,
                     MymodError_t *status)
{
   int stat;
   MymodError_t err = kMymodErr_Success;
   int nSets = 0;
   int iSet = 0;
   int nRecs = 0;
   int iRec = 0;
   unsigned long long tsize = 0;
   DRMS_Record_t *recout = NULL;
   DRMS_Record_t *recin = NULL;
   DRMS_RecordSet_t *rs = NULL;
   DRMS_RecordSet_t *rsout = NULL;
   DRMS_RecordSet_t *rsin = NULL;
   char rsoutquery[DRMS_MAXQUERYLEN];
   char *rsinquery = NULL;

   snprintf(rsoutquery, sizeof(rsoutquery), "%s[%s]", expseries, reqid);

   rs = drms_open_records(env, rsoutquery, &stat);

   if (rs)
   {
      /* XXX Change to DRMS_REPLACE_SEGMENTS (which isn't implemented yet). */
      rsout = drms_clone_records(rs, DRMS_PERMANENT, DRMS_COPY_SEGMENTS, &stat);
      drms_close_records(rs, DRMS_FREE_RECORD);
   }
   
   if (rsout && rsout->n == 1)
   {
      /* Filename Format comes from export series */
      char *ffmt = NULL;
      char *kval = NULL;

      recout = rsout->records[0];
      rsinquery = drms_getkey_string(recout, drms_defs_getval("kExportKW_Request"), &stat);
      
      kval = drms_getkey_string(recout, drms_defs_getval("kExportKW_FileNameFormat"), &stat);
      if (kval)
      {
         if (*kval == '\0')
         {
            ffmt = NULL;
         }
         else
         {
            ffmt = strdup(kval);
         }

         free(kval);
      }

      if (rsinquery && (rsin = drms_open_records(env, rsinquery, &stat)))
      {
         /* stage records to reduce number of calls to SUMS. */
         drms_stage_records(rsin, 1, 0);
	 nSets = rsin->ss_n;

	 for (iSet = 0; 
              stat == DRMS_SUCCESS && err == kMymodErr_Success && iSet < nSets; 
              iSet++)
	 {
	    /* Perhaps we will need this in the future?
	     * request = rsin->ss_queries[iSet];
	     */
	    nRecs = drms_recordset_getssnrecs(rsin, iSet, &stat);

	    for (iRec = 0; 
                 stat == DRMS_SUCCESS && err == kMymodErr_Success &&iRec < nRecs; 
                 iRec++)
	    {
	       recin = rsin->records[(rsin->ss_starts)[iSet] + iRec];
	       tsize += MapexportRecord(recout, 
                                        recin,
                                        classname, 
                                        mapfile, 
                                        pklfilename,
                                        tcount, 
                                        ffmt, 
                                        outpath, 
                                        pklist, 
                                        cparms, 
                                        &err);
	    }
	 }
	 
	 if (stat != DRMS_SUCCESS || err != kMymodErr_Success)
	 {
	    fprintf(stderr, "Export halted due to DRMS failure.\n");
	 }
         else if (exptime)
         {
            *exptime = CURRENT_SYSTEM_TIME;
         }
      }
      else
      {
	 fprintf(stderr, 
		 "Export series keyword '%s' did not contain a valid recordset query.\n", 
                 drms_defs_getval("kExportKW_Request"));
         err = kMymodErr_BadRecSetQuery; 
      }

      if (ffmt)
      {
         free(ffmt);
      }
   }
   else
   {
      fprintf(stderr, "Could not open export destination record set with request id '%s'.\n", reqid);
      err = kMymodErr_BadRequestID; 
   }

   /* Set output record keywords. */
   if (err == kMymodErr_Success)
   {
      drms_setkey_time(recout, drms_defs_getval("kExportKW_ExpTime"), CURRENT_SYSTEM_TIME);
      drms_setkey_int(recout, drms_defs_getval("kExportKW_DataSize"), (int)ToMB(tsize));
   }

   if (rsout)
   {
      drms_close_records(rsout, DRMS_INSERT_RECORD);
   }

   if (rsinquery)
   {
      free(rsinquery);
   }

   if (status)
   {
      *status = err;
   }
   
   return tsize;
}

static char *GenErrMsg(const char *fmt, ...)
{
   char *msgout = NULL;
   char errmsg[4096];

   va_list ap;
   va_start(ap, fmt);
   vsnprintf(errmsg, sizeof(errmsg), fmt, ap);

   msgout = strdup(errmsg);
   fprintf(stderr, errmsg);
   
   va_end (ap);

   return msgout;
}

static MymodError_t AppendContent(FILE *dst, FILE *src)
{
   MymodError_t err = kMymodErr_Success;
   char buf[32768];
   int nread = 0;
   int dstfd;
   int srcfd;

   if (dst && src)
   {
      dstfd = fileno(dst);
      srcfd = fileno(src);

      while(1)
      {
         nread = read(srcfd, buf, sizeof(buf));
         if (nread > 0)
         {
            if (write(dstfd, buf, nread) == -1)
            {
               err = kMymodErr_PackfileFailure;
               fprintf(stderr, "Failure writing packing-list file.\n");
               break;
            }
         }
         else
         {
             if (nread == -1)
             {
                err = kMymodErr_PackfileFailure;
                fprintf(stderr, "Failure reading packing-list file.\n");
             }

             break;
         }
      }
   }

   return err;
}

//#undef DEFS_MKPATH
//#define FRED(X, Y) X##Y
//#define DEFS_MKPATH(Y) FRED(CDIR, Y)
//#define DEFS_MKPATH(Y) CDIR##Y

int DoIt(void) 
{
   MymodError_t err = kMymodErr_Success;
   int drmsstat = DRMS_SUCCESS;
   long long tsize = 0; /* total size of export payload in bytes */
   long long tsizeMB = 0; /* total size of export payload in Mbytes */
   int tcount = 0;
   TIME exptime = DRMS_MISSING_TIME;
   FILE *pklist = NULL;
   FILE *pklistTMP = NULL;
   char pklistfname[PATH_MAX];
   char pklistfnameTMP[PATH_MAX];
   char pklistpath[PATH_MAX];
   char pklistpathTMP[PATH_MAX];

   const char *version = NULL;
   const char *reqid = NULL;
   const char *method = NULL;
   //ISS fly-tar START
   const char *ackfile= NULL;
   const char *tarfile= NULL;
   const char *segments= NULL;
   //ISS fly-tar END
   const char *protocol = NULL;
   const char *rsquery = NULL;
   const char *clname = NULL;
   const char *mapfile = NULL;
   const char *cparmsarg = NULL;
   const char **cparms = NULL;
   int RecordLimit = 0;

   /* "packing list" header/metadata */
   char *md_version = NULL;
   char *md_reqid = NULL;
   char *md_method = NULL;
   char *md_protocol = NULL;
   char *md_count = NULL; /* number of files exported */
   char *md_size = NULL; /* total Mbytes exported */
   char *md_exptime = NULL;
   char *md_dir = NULL; /* same as outpath, the /SUMX path that contains the exported files
                         * before they are downloaded by the user */
   char *md_status = NULL; 
   char *md_error = NULL;

   RecordLimit = cmdparams_get_int(&cmdparams, kArg_n, &drmsstat);

   if (drms_defs_register(DEFS_MKPATH("/data/export.defs")))
   {
      md_error = GenErrMsg("jsoc_export failure - missing definition file.\n");
      err = kMymodErr_CantRegisterDefs;
   }
   else
   {
      snprintf(pklistfname, sizeof(pklistfname), "%s", drms_defs_getval("kPackListFileName"));
      snprintf(pklistfnameTMP, sizeof(pklistfnameTMP), "%s.tmp", pklistfname);

      version = cmdparams_get_str(&cmdparams, kArg_version, &drmsstat);
      reqid = cmdparams_get_str(&cmdparams, kArg_reqid, &drmsstat);
      method = cmdparams_get_str(&cmdparams, kArg_method, &drmsstat);

      //ISS fly-tar START copy the argument to a static variable. 
      ackfile= cmdparams_get_str(&cmdparams, kArg_ackfile, &drmsstat);
      if (strcmp(ackfile,kNotSpecified) && access(ackfile, F_OK) == 0 ) {
        fprintf(stderr, "Including ACK file [%s]\n",ackfile);
        strcpy(sackfile,ackfile);
      }

      tarfile=cmdparams_get_str(&cmdparams, kArg_tarfile, &drmsstat);
      if (strcmp(tarfile,kNotSpecified)) {
        fprintf(stderr, "Including tar file name [%s]\n",tarfile);
        strcpy(starfile,tarfile);
      }

      segments=cmdparams_get_str(&cmdparams, kArg_segments, &drmsstat);

      //ISS fly-tar END copy the argument to a static variable. 

      protocol = cmdparams_get_str(&cmdparams, kArg_protocol, &drmsstat);

      rsquery = cmdparams_get_str(&cmdparams, kArg_rsquery, &drmsstat);

      clname = cmdparams_get_str(&cmdparams, kArg_clname, &drmsstat);
      if (drmsstat != DRMS_SUCCESS || !strcmp(clname, kNotSpecified))
      {
         clname = NULL;
      }

      mapfile = cmdparams_get_str(&cmdparams, kArg_kmfile, &drmsstat);
      if (drmsstat != DRMS_SUCCESS || !strcmp(mapfile, kNotSpecified))
      {
         mapfile = NULL;
      }

      cparmsarg = cmdparams_get_str(&cmdparams, kArg_cparms, &drmsstat);
      if (strcmp(cparmsarg, kNotSpecified))
      {
         char *dup = strdup(cparmsarg);
         char *pc = NULL;
         char *pend = NULL;
         int nstr;
         int istr;
         
         /* count number of compression strings (one for each segment being exported) */
         pc = dup;
         nstr = 1;
         while ((pc = strchr(pc, ',')) != NULL)
         {
            pc++;
            ++nstr;
         }

         cparms = (const char **)malloc((nstr + 1) * sizeof(char *));
         
         pc = dup;
         for (istr = 0; istr < nstr; istr++)
         {
            pend = strchr(pc, ',');
            if (pend)
            {
               *pend = '\0';
            }
            cparms[istr] = (strcmp(pc, kNoCompression) == 0) ? strdup("") : strdup(pc);
            if (pend)
            {
               pc = pend + 1;
            }
         }

         /* Empty string to indicate end */
         cparms[nstr] = NULL;

         if (dup)
         {
            free(dup);
         }
      }

      md_version = strdup(version);
      md_reqid = strdup(reqid);
      md_method = strdup(method);
      md_protocol = strdup(protocol);

      if (strcmp(rsquery, kNotSpecified) == 0)
      {
         /* The record-set query that identifies records to export comes from 
          * the export series, not from the command-line. A dbase query for the
          * reqid in the export series returns a record which contains the
          * record-set query. This export-series record is cloned, and the 
          * output files are written to the clone's SUDIR. 
          *
          * The output filename format is obtained from the export series record */
         char *outpath = NULL;

         /* No record-set query provided - must use series to get rsquery */
         const char *expseries = NULL;

         expseries = cmdparams_get_str(&cmdparams, kArg_expSeries, &drmsstat);

         /* Open packing-list file */         
         tsize = Mapexport(drms_env, 
                           reqid, 
                           clname, 
                           mapfile, 
                           expseries, 
                           pklistfnameTMP, 
                           &tcount, 
                           &outpath, 
                           &exptime, 
                           &pklistTMP,
                           cparms, 
                           &err);

         if (err != kMymodErr_Success)
         {
            md_error = GenErrMsg("Failure occurred while processing export Request ID '%s'.\n", reqid);
            err = kMymodErr_ExportFailed;
         }
         else
         {
            snprintf(pklistpathTMP, sizeof(pklistpathTMP), "%s/%s", outpath, pklistfnameTMP);

            /* Set packing list info */
            md_dir = strdup(outpath);
         }

         if (outpath)
         {
            free(outpath);
         }
      }
      else
      {
         /* Packing list items that come from the cmd-line arguments. */
         const char *outpath = NULL;

         /* Filename Format comes from cmd-line argument */
         const char *ffmt = NULL;
      
         outpath = cmdparams_get_str(&cmdparams, kArg_path, &drmsstat);
         //ISS fly-tar only get in if it is not cgi
         if (strcmp(outpath, kNotSpecified) == 0 && strcmp(method,"url_cgi"))
         {
            /* Use current working directory by default */
            outpath = getenv(kPWD);
         }

         ffmt = cmdparams_get_str(&cmdparams, kArg_ffmt, &drmsstat);
         if (strcmp(ffmt, kNotSpecified) == 0 || *ffmt == '\0')
         {
            /* Assume the user wants the default filename format - set to NULL */
            ffmt = NULL;
         }

//ISS fly-tar start
         if (strcmp(method,"url_cgi")) {
           /* Open tmp packing-list file */
           snprintf(pklistpathTMP, sizeof(pklistpathTMP), "%s/%s", outpath, pklistfnameTMP);
           pklistTMP = fopen(pklistpathTMP, "w+");
         } else {  //if method is "url_cgi" we don't want tmp file 
           pklistTMP=NULL;
         }
//ISS fly-tar end
         /* Open tmp packing-list file */

         if (pklistTMP || !strcmp(method,"url_cgi")) //ISS fly-tar
         {
            /* Call export code, filling in tsize, tcount, and exptime */
            tcount = RecordLimit;
            tsize = MapexportToDir(drms_env, 
                                   rsquery, 
                                   ffmt, 
                                   outpath, 
                                   method, //ISS fly-tar
                                   pklistTMP, 
                                   clname, 
                                   mapfile, 
                                   &tcount, 
                                   &exptime, 
                                   cparms, 
                                   &err);
         }
         else
         {
            err = kMymodErr_CantOpenPackfile;
            md_error = GenErrMsg("Couldn't open packing-list file '%s'.\n",  pklistpathTMP);
         }

         if (err != kMymodErr_Success)
         {
            md_error = GenErrMsg("Failure occurred while processing export Request ID '%s'.\n", reqid);
            err = kMymodErr_ExportFailed;
         }
         else
         {
            /* Set packing list info */
            md_dir = strdup(outpath);
         }
      }

      tsizeMB = ToMB(tsize);

      if (drmsstat != DRMS_SUCCESS)
      {
         md_error = GenErrMsg("DRMS error '%d'.\n", drmsstat);
      }
      else if (err == kMymodErr_Success)
      {
         char tstr[64];
         int strsize = 0;
         sprint_time(tstr, exptime, "UT", 0);

         /* Set packing list info not already set */
         strsize = 64;
         md_size = malloc(strsize);
         snprintf(md_size, strsize, "%lld", tsizeMB);
         md_count = malloc(strsize);
         snprintf(md_count, strsize, "%d", tcount);
         md_exptime = strdup(tstr);
      }

      if (strcmp(method,"url_cgi") != 0)
        fprintf(stdout, "%lld megabytes exported.\n", tsizeMB); 
   }

   /* open 'real' pack list */
   if (md_dir && strcmp(method,"url_cgi")) //ISS fly-tar 
   {
      snprintf(pklistpath, sizeof(pklistpath), "%s/%s", md_dir, pklistfname);
      pklist = fopen(pklistpath, "w+");
   }

   if (pklist && strcmp(method,"url_cgi")) //ISS fly-tar
   {
      if (fseek(pklistTMP, 0, SEEK_SET))
      {
         md_error = GenErrMsg("Failure accessing packing-list file '%s'.\n", pklistfnameTMP);
         err = kMymodErr_PackfileFailure;
      }
   }
   else if (strcmp(method,"url_cgi")) //ISS fly-tar
   {
      md_error = GenErrMsg("Failure opening packing-list file '%s'.\n", pklistfname);
      err = kMymodErr_PackfileFailure;
   }

   /* For now, there is just success/failure in the packing-list file */
   if (err == kMymodErr_Success)
   {
      md_status = strdup(drms_defs_getval("kMDStatus_Good"));
   }
   else
   {
      md_status = strdup(drms_defs_getval("kMDStatus_Bad"));
   }

   if (pklist)
   {
      /* write packing list */
      fprintf(pklist, "# JSOC \n");
      WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_Version"), md_version);
      WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_RequestID"), md_reqid);
      WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_Method"), md_method);
      WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_Protocol"), md_protocol);
      WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_Count"), md_count ? md_count : "-1");
      WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_Size"), md_size ? md_size : "-1");
      WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_ExpTime"), md_exptime ? md_exptime : "");
      WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_Dir"), md_dir ? md_dir : "");
      WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_Status"), md_status);

      fflush(pklist);

      if (err == kMymodErr_Success)
      {
         /* Copy the content from the temporary pklist into the real pklist */
         fprintf(pklist, "# DATA \n");
         fflush(pklist);
         err = AppendContent(pklist, pklistTMP);
      }

      if (err != kMymodErr_Success)
      {
         fflush(pklist);

         /* If there is an error, write out the error message. */
         WritePListRecord(kPL_metadata, pklist, drms_defs_getval("kMD_Error"), md_error ? md_error : "");
      }
   }



   if (pklistTMP)
   {
      /* close tmp packing-list file */
      fclose(pklistTMP);
      pklistTMP = NULL;

      /* delete temporary file */
      unlink(pklistpathTMP);
   }

   if (pklist)
   {
      /* close actual packing-list file */
      fclose(pklist);
      pklist = NULL;
   }

   if (md_version)
   {
      free(md_version);
   }
   if (md_reqid)
   {
      free(md_reqid);
   }
   if (md_method)
   {
      free(md_method);
   }
   if (md_protocol)
   {
      free(md_protocol);
   }
   if (md_count)
   {
      free(md_count);
   }
   if (md_size)
   {
      free(md_size);
   }
   if (md_exptime)
   {
      free(md_exptime);
   }
   if (md_dir)
   {
      free(md_dir);
   }
   if (md_status)
   {
      free(md_status);
   }
   if (md_error)
   {
      free(md_error);
   }

   if (cparms)
   {
      int iseg = 0;
      while (cparms[iseg])
      {
         free((void *)cparms[iseg]);
         iseg++;
      }

      free(cparms);
   }
   
   return err;
}

//ISS fly-tar START
int flytar_append_file_header(TAR *t, char * filename, int size) {
   //setting header block
   memset(&(t->th_buf), 0, sizeof(struct tar_header));
   file_set_stats(t, size, filename);
   fflush(NULL);
   th_write(t);
   fflush(NULL);
   return 0;
}

int flytar_append_file_tail(TAR *t, int size) {
   char block[512];

   memset(block, 0, 512);
//   fprintf(stderr,"block size to add [%d]\n", block_rest(512,size));
   fflush(NULL);
   if (-1==(*((t)->type->writefunc))((t)->fd, block, block_rest(512,size))) {
     return -1;
   }
   fflush(NULL);
   return 0;
}

int flytar_append_file(TAR *t, char * fullname) {
  struct stat   s;
  int           ACK        =  -1;
  int           tsize      =   0; 
  if (lstat(fullname,&s) != 0)
    return 1;

  char *basename = strrchr(fullname, '/');
  basename = basename? basename + 1: (char *) fullname;
//  fprintf(stderr, "Adding ACK file [%s] ; basename [%s] size [%d]\n", fullname, basename, s.st_size);
  flytar_append_file_header(t, basename, s.st_size);


  if ( (ACK=open(fullname,O_RDONLY))>-1  ) {
    char buffer[81];
    int  rest=0;
    while (( rest = read(ACK,buffer, 80)) == 80) {
      tsize += 80;
      buffer[80]='\0';
      fprintf(stdout,"%s",buffer);
    }
    if (rest > 0 ) {
      tsize += rest;
      read(ACK,buffer,rest);
      buffer[rest]='\0';
    }

    fprintf(stdout,"%s",buffer);

    close(ACK);

  } else {
    return 1;
  }

//  fprintf(stderr, "Before padding ACK file size [%d] and read size [%d]\n", s.st_size, tsize);
  flytar_append_file_tail(t,s.st_size);

  return 0;
}

int block_rest(int block_size, int file_size) {
  int mod = 0;
  if ( file_size < block_size) {
   return block_size - file_size;
  } else {
   mod = file_size%block_size;
   if (mod > 0) {
     return 512 - mod;
   } else {
     return mod;
   }
  }
}

int file_set_stats(TAR *t, int size, char * filename) {
    t->th_buf.typeflag = REGTYPE;

    //setting user info
    strcpy(t->th_buf.uname, "production");
	int_to_oct(670, t->th_buf.uid, 8);

    //setting group info
    strcpy(t->th_buf.gname, "SUMS");
	int_to_oct(669, t->th_buf.gid, 8);

    //setting file mode
    int_to_oct(33204,t->th_buf.mode,8);

    //setting time
    int_to_oct(time(NULL),t->th_buf.mtime,12);

    //setting size
    int_to_oct(size,t->th_buf.size,12);

    //setting pathname
    th_set_path(t,filename);
}

char * replace_blanks(char *str1, const char *str2, int n) {
//  char *query ="lm_jps.lev1_test4k10s[?t_obs_index between 542412329 and 542412424 and wave_str='171_THIN'?]";
  char *str2cpy   = NULL;
  char *ptr        = NULL;

  ptr=str2cpy=malloc(sizeof(char)*(strlen(str2) +1));
  strcpy(str2cpy,str2);

  while ( *ptr != '\0' ) {

    if (*ptr==' ' || *ptr=='[' || *ptr == ']'  || *ptr == '?' || *ptr == '\'' || *ptr == '=' || *ptr == '>' || *ptr =='<' || *ptr == '{'
        || *ptr=='}')
      *ptr='_';

    ptr++;
  }

  strncpy(str1,str2cpy,n);
  free(str2cpy);
  return str1;
}


int match_query_segments(DRMS_Segment_t *seg)
{
  int j=0;
  if (segments!=NULL) {
    for (j=0; segments[j][0] !='\0'; j++) {
/*
      fprintf(stderr,"segment name [%s]\n",seg->info->name);
      fprintf(stderr,"segment counter [%d]\n",j);
      fprintf(stderr,"segment pass name [%s]\n",segments[j]);
*/
      if (strcmp(segments[j],seg->info->name) == 0) {
        return 0;
      }
    }
    return 1;
  } else {
    return 0;
  }
}
int count_segment_names(char **segments)
{
  int j=0;
  while(segments!=NULL && segments[++j][0]!='\0');
  return j;
}

char ** parse_query_for_segment_names(char * query)
{
  char **segments = (char **) NULL;

  char *cquery = (char *) NULL;

  char *ptr1, *ptr2;
  //duplicate query
  cquery=strdup(query);

  //find begin and end brackets
  ptr1 = strchr( cquery, '{');
  ptr2 = strrchr(cquery, '}');
  if (ptr1 != NULL && ptr2 != NULL) {
    *ptr2='\0';  //overwrite end bracket with str teminator
    ++ptr1;      //Position pointer in first char after first bracket

    char *segstr=ptr1;  // This is the full segment list string

    //count commas
    int commaCount=0;
    while ((ptr2 = strchr( ptr1, ',')) != NULL) {
      ptr1=ptr2+1;
      commaCount++;
    }

    //the segment array has always commaCount + 2 elements. 
    //where the last element is always null;
    int arrCount = commaCount+2;
    segments = malloc(arrCount*sizeof(char *));

    //add last element in array; always an empty string.   
    segments[arrCount-1] = malloc(sizeof(char));
    segments[arrCount-1][0]='\0';

    //reuse ptr1 
    ptr1 = segstr;
    int i = 0;
    int segNameLen = 0;
    //find the segment names and populate array
    while((ptr2 = strchr( ptr1, ',')) != NULL) {
       segNameLen= ptr2 - ptr1;
       segments[i]=malloc(segNameLen*sizeof(char)+1);
       strncpy(segments[i],ptr1,segNameLen);
       segments[i][segNameLen]='\0';
       ptr1=ptr2+1;
       i++;
    }

    segNameLen=strlen(ptr1);
    segments[i]=malloc(segNameLen*sizeof(char)+1);
    strcpy(segments[i],ptr1);
    segments[i][segNameLen]='\0';

    //test the array
/*
    int j=0;
    while(segments !=NULL && segments[j++][0]!='\0') {
      fprintf(stderr,"in parser, segment name [%s]\n", segments[j-1]);
    }
*/
    free(cquery);
  }
  return segments;
}

int testRecordCount(DRMS_RecordSet_t* rsin, int* stat, int querySegCount)
{
   MymodError_t modstat = kMymodErr_Success;
   DRMS_Record_t *recin = NULL;
   int iSet =0, nSets = 0, nRecs = 0, recCount = 0, iRec = 0, segNum = 0;   

   nSets = rsin->ss_n; //number of Sets?

   //test for multiple records (fly-tar mode)
   for (iSet = 0; 
           *stat == DRMS_SUCCESS && modstat == kMymodErr_Success && iSet < nSets; 
           iSet++)
      {
         nRecs = drms_recordset_getssnrecs(rsin, iSet, stat);
         //fprintf(stderr, "testRecordCount nRecs[%d]",nRecs);
         for (iRec = 0; modstat == kMymodErr_Success && iRec < nRecs; iRec++)
         {
           recin =rsin->records[(rsin->ss_starts)[iSet] + iRec];
           //fprintf(stderr, "testRecordCount recCount[%d]",recCount);
           if ( querySegCount != 1) {
             if ( (segNum = drms_record_numsegments(recin)) > 1) {
               recCount=segNum;
             }
           }
           recCount ++;
         }
      }

   return recCount;
}

void CallbackErrorBuffer( char* filename, char*query, int errorCode) {
  char line[4096];
  char error[4096];

  if ( 2 == errorCode) {
    strcpy(error,"###Tar file too big. Please reduced your query range. Max tar file 2GB. Contact igor@noao.edu or oneiros@grace.nascom.nasa.gov if you need help #\n");
  } else {
    strcpy(error,"###The data requested could not be served. Please retry again in a few minutes. If the problem persists please contact: igor@noao.edu or oneiros@grace.nascom.nasa.gov  #\n");
  }

  snprintf(line, 4095, "%s; Query=[%s]\n",filename,query);

  int linelen = strlen(line);
  char *ptr = errorbuffer;

  if (errorbuffer == (char *) NULL) {

    int      errorbufferlen     = strlen(error) + linelen + 1;  // 1 is for EndOfString char \0


    errorbuffer =  malloc(sizeof(char)*errorbufferlen);
    fprintf(stderr,"Allocating errorbuffer [%x]\n",errorbuffer);

    memset(errorbuffer,0,sizeof(char)*errorbufferlen);

    sprintf(errorbuffer,"%s%s",error,line);
  } else {
    int currentbufferlen = strlen(errorbuffer) + 1; // 1 is for EndOfString char \0

    fprintf(stderr,"Reallocating errorbuffer [%x] current buffer lenght [%d] and addition is [%d] and [%d]\n",errorbuffer, currentbufferlen,linelen, strlen(error));
    errorbuffer = (char *) realloc(errorbuffer, sizeof(char)*(currentbufferlen + linelen + strlen(error)));

    //repointing errorbuffer to unwritten space
    ptr= errorbuffer + strlen(errorbuffer)*sizeof(char);

    //memset newly allocated memory.
    memset(ptr,0,sizeof(char)*(linelen+strlen(error)));
    sprintf(ptr,"%s%s",error,line);
  }

}

void CallbackClose(const char *rsquery, TAR **tar2, int recCount) {
  TAR *tar = *tar2;

//  fprintf(stderr, "closing tar, records [%d] tar ptr [%x]\n",recCount,tar); 
  if (recCount > 1 && tar != (TAR *) NULL) {
//    fprintf(stderr,"Closing tar file\n");
    if (!print_header) {
      char tarfilename[300];
      char starfile[300];
      printf("Content-type: application/octet-stream\n");
      printf("Content-Disposition: attachment; filename=\"%s.tar\"\n",strlen(starfile)==0?replace_blanks(tarfilename,rsquery,300):starfile);
      printf("Content-transfer-encoding: binary\n\n");
      print_header = 1;
    }
    fflush(NULL);
    if (infobuffer !=NULL) {
      if (flytar_append_file_header(tar, "file_list.txt", strlen(infobuffer))){
        fprintf(stderr, "Error in flytar_append_file\n");
      } else {
  //                 fprintf(stderr,"infobuffer [%s]\n",infobuffer);
        fprintf(stdout,infobuffer);
        if (flytar_append_file_tail(tar, strlen(infobuffer))) {
          fprintf(stderr, "Error in flytar_append_tail\n");
        }
        free(infobuffer);
      }
    }
  
    fflush(NULL);
    if ((errorbuffer != (char *) NULL)) {
      if (flytar_append_file_header(tar, "error_list.txt", strlen(errorbuffer))){
        fprintf(stderr, "Error in flytar_append_file\n");
      } else {
        fprintf(stdout,errorbuffer);
        if (flytar_append_file_tail(tar, strlen(errorbuffer))) {
          fprintf(stderr, "Error in flytar_append_tail\n");
        }
        free(errorbuffer);
      }
    }
              
    //fprintf(stderr,"before appending ACK file\n");
    fflush(NULL);
    flytar_append_file(tar,sackfile);
    fprintf(stderr,"tar closing ... \n");
    tar_append_eof(tar);
    tar_close(tar);
    *tar2 = NULL;
//    fprintf(stderr,"tar pointer [%x]\n",tar);
    fflush(NULL);
  } else {
    if (errorbuffer != NULL) {
        //char *basename = strrchr(fitsfilename, '/');
        //basename = basename? basename + 1: (char *) fitsfilename;
        if (!print_header) {
          printf("Content-type: text/html\n\n");
          print_header = 1;
        }
        printf("<html><base>\n<h1>Retrieval for the below data failed</h1>\n");
        //printf("<p>DRMS query:<font color=\"blue\">[%s]</font></p>\n",rsquery);
        printf("<p><i><font color=\"blue\">%s</font></i></p>\n",errorbuffer);
        printf("<p>If a retry does not fix the problem, it could be the data is missing. In which case, as there are no data regeneration capabilities at JSOC at the moment, it should be considered as missing data and not retrievable at the present time.</p>\n<p>We are currently working on showing the detailed reason for the export failure.</p>\n");
        printf("<p>Thanks for your patience</p>\n<p>Please contact <a href=\"mailto:igor@noao.edu\">igor@noao.edu</a> or  <a href=\"mailto:oneiros@grace.nascom.nasa.gov\">oneiros@grace.nascom.nasa.gov</a> for further questions.</p>\n");
        printf("</base></html>");
//        errorbuffer=NULL;
    }
  } 
}

void CallbackInfo(char * filename) {
  char line[400];

  sprintf(line,"%s\n",filename);

  int linelen = strlen(line);
  char *ptr = infobuffer;


  if (infobuffer == (char *) NULL) {

    char    *infohead          = "# List of files in Tar file #\n";
    int      infobufferlen     = strlen(infohead) + linelen + 1;  // 1 is for EndOfString char \0

    infobuffer =  malloc(sizeof(char)*infobufferlen);

    memset(infobuffer,0,sizeof(char)*infobufferlen);

    sprintf(infobuffer,"%s",infohead);
    //repointing infobuffer to unwritten space
    ptr = infobuffer + strlen(infohead)*(sizeof(char));

  } else {
    int currentbufferlen = strlen(infobuffer) + 1; // 1 is for EndOfString char \0

    infobuffer = (char *) realloc(infobuffer, sizeof(char)*(currentbufferlen + linelen));

    //repointing infobuffer to unwritten space
    ptr= infobuffer + strlen(infobuffer)*sizeof(char);

    //memset newly allocated memory.
    memset(ptr,0,sizeof(char)*linelen);
  }

  strcpy(ptr,line);

}
//ISS fly-tar END
