// #define DEBUG 1
#define DEBUG 0

// insubstantial change to test make

// TESTMODE=1 means look for status=12 in jsoc.export_new, used for testing jsoc_export_manage
// #define TESTMODE 1
#define TESTMODE 0

/* 
 * ART - BUG
 * The actual size of export is never saved in the Size keyword of the jsoc.export series. jsoc_fetch
 * writes the size that SUM_info returns into the Size keyword of the jsoc.export_new series, 
 * which is then copied to jsoc.export.  But this value is meaningless since the export hasn't even
 * been performed yet, and I'm not even sure exactly what the value returned by SUM_info means.
 * We need to somehow calculate the size of the export here and write it to the Size keyword
 * of jsoc.export. The value does exist in the index.txt file (which is generated by jsoc_export_as_fits
 * at least). jsoc_export_as_fits cannot directly write the value because it does not have access
 * to the export record in jsoc.export.
 */

/*
 *  jsoc_export_manage - program to manage jsoc export requests posted to jsoc.export
  
  This program provides some service functions to manage export requests using
  the jsoc.export, jsoc.export_new, and jsoc.export_user series.

  In the default mode the program watches for new requests and submits them to
  the cluster for processing.  The submission is handled via a pair of scripts.
  the first is a qsub script that executes a drms_run call of the second.
  The steps are as follows:

    1.  foreach record with status==2 in jsoc.export_new do: 
        a. create new record in jsoc.export and fill with contents of
           the jsoc.export_new record.
        b. perhaps later handle the contact info better to work with
           a list of users and addresses.  For now just leave jsoc_export_user
           alone.
        c. get record directory of new jsoc.export record, REQDIR.
        c. get the Processing keyword and check for valid operation.
        d. Get the DataSet keyword
        e. Build processing command 
        f. get RequestID
        g. get other keywords such as Notify.
        g. create a script file at: REQDIR/<RequestID>.qsub that contains:
                # /bin/csh -f
		cd REQDIR
                drms_run <RequestID>.drmsrun
                if ($status) then
                  set_keys ds="jsoc.export[<RequestID>] status=4
		  set subject="JSOC export <RequestID> FAILED"
                  set msg="Error status returned from DRMS session, see log files at http://jsoc.stanford.edu/REQDIR"
                  mail -s "JSOC export <RequestID> FAILED" <NOTIFY> <<!
                  Error status returned from DRMS session.
		  See log files at http://jsoc.stanford.edu/REQDIR
                  !
                endif
                
        h. create a script file at: REQDIR/<RequestID>.drmsrun
            1. write preamble for running in cluster via qsub.
            2. add processing command with stdout and stderr to local files in REQDIR
            3. add generation of index.html, index.txt, index.json, index.xml(?)
            4. set status=0 in jsoc.export[<RequestID>]
        i  perhaps add initial versions of index.html, etc. 
        j. execute the program: qsub -j.q REQDIR/<RequestID>.qsub
        k. set status=1 in jsoc.export[<RequestID>] in a (maybe)transient record.
     
  Later additions to this program may provide monitoring services to watch the status.

  Command line args of JSOC_DBHOST and JSOC_DBMAINHOST are now used to direct queries to
  one database server for export processing (JSOC_DBHOST) and another for access to science
  data to be exported (JSOC_DBMAINHOST).

 *
 */
#include "jsoc_main.h"
#include "drms.h"
#include "drms_names.h"
#include "json.h"
#include "serverdefs.h"

#define EXPORT_SERIES "jsoc.export"
#define EXPORT_SERIES_NEW "jsoc.export_new"
#define EXPORT_USER "jsoc.export_user"

#define PACKLIST_VER "0.5"

#define kArgTestmode    "t"

#define DIE(msg) { fprintf(stderr,"XXXX jsoc_exports_manager failure: %s\nstatus=%d",msg,status); exit(1); }

#define kHgPatchLog "hg_patch.log"

enum Protocol_enum
{
   kProto_AsIs = 0,
   kProto_FITS = 1,
   kProto_MPEG = 2,
   kProto_JPEG = 3,
   kProto_PNG  = 4
};

typedef enum Protocol_enum Protocol_t;

const char *protos[] =
{
   "as-is",
   "fits",
   "mpeg", 
   "jpeg",
   "png"
};

enum Processing_enum
{
   kProc_Unk = 0,
   kProc_Noop = 1,
   kProc_NotSpec = 2,
   kProc_HgPatch = 3,
   kProc_SuExport = 4
};

typedef enum Processing_enum Processing_t;

const char *procs[] =
{
   "uknown",
   "no_op",
   "Not Specified",
   "hg_patch",
   "su_export"
};

ModuleArgs_t module_args[] =
{ 
  {ARG_STRING, "op", "process", "<Operation>"},
  {ARG_FLAG, kArgTestmode, NULL, "if set, then operates on new requests with status 12 (not 2)"},
  {ARG_FLAG, "h", "0", "help - show usage"},
  {ARG_END}
};

char *module_name = "jsoc_export_manage";

int nice_intro ()
  {
  int usage = cmdparams_get_int (&cmdparams, "h", NULL);
  if (usage)
    {
    printf ("Usage:\njsoc_info {-h} "
        "  details are:\n"
	"op=<command> tell which function to execute\n"
        "   choices are:\n"
        "       process - get new export requests and submit scripts.\n"
        "       TBD\n"
	"h=help - show usage\n"
	);
    return(1);
    }
  return (0);
  }

// generate qsub script
void make_qsub_call(char *requestid, char *reqdir, int requestorid, const char *dbname, 
               const char *dbuser, const char *dbids, const char *dbexporthost, const char *dbmainhost)
  {
  FILE *fp;
  char qsubscript[DRMS_MAXPATHLEN];
  // Note that the reqdir where the script is put and is run from will not
  // be the record dir that is visible when the drms_run run is finished.
  sprintf(qsubscript, "%s/%s.qsub", reqdir, requestid);
  fp = fopen(qsubscript, "w");
  fprintf(fp, "#! /bin/csh -f\n");
  fprintf(fp, "set echo\n");
  // fprintf(fp, "setenv SUMSERVER d02\n");
  // fprintf(fp, "unsetenv DRMSSESSION\n");
  // fprintf(fp, "unsetenv SETJSOCENV\n");
  // fprintf(fp, "unsetenv SET_SOLAR_ENV\n");
  // fprintf(fp, "source /home/phil/.sunrc\n");
  // fprintf(fp, "source /home/jsoc/.setJSOCenv\n");
  // fprintf(fp, "printenv\n");

  /* Set path based on export root (if set) */
  fprintf(fp, "if (${?JSOCROOT_EXPORT}) then\n");
  fprintf(fp, "  set path = ($JSOCROOT_EXPORT/bin/$JSOC_MACHINE $JSOCROOT_EXPORT/scripts $path)\n");
  fprintf(fp,"endif\n");

  fprintf(fp, "while (`show_info JSOC_DBHOST=%s -q 'jsoc.export_new[%s]' key=Status %s` == 2)\n", dbexporthost, requestid, dbids);
  fprintf(fp, "  echo waiting for jsocdb commit >> /home/jsoc/exports/tmp/%s.runlog \n",requestid);
  fprintf(fp, "  sleep 1\nend \n");
  if (dbname)
  {
     fprintf(fp,   "setenv JSOC_DBNAME %s\n", dbname);
  }
  if (dbuser)
  {
     fprintf(fp,   "setenv JSOC_DBUSER %s\n", dbuser);
  }
// TEST
//fprintf(fp,   "setenv JSOC_DBHOST %s\n", dbmainhost);
  fprintf(fp,   "setenv JSOC_DBHOST %s\n", dbexporthost);
  fprintf(fp,   "setenv JSOC_DBEXPORTHOST %s\n", dbexporthost);
  fprintf(fp, "drms_run JSOC_DBHOST=%s %s/%s.drmsrun >>& /home/jsoc/exports/tmp/%s.runlog \n", dbexporthost, reqdir, requestid, requestid);
  fprintf(fp, "set DRMS_ERROR=$status\n");
  fprintf(fp, "set NewRecnum=`cat /home/jsoc/exports/tmp/%s.recnum` \n", requestid);
  fprintf(fp, "set WAITCOUNT = 60\n");
  fprintf(fp, "while (`show_info JSOC_DBHOST=%s -q -r 'jsoc.export[%s]' %s` < $NewRecnum)\n", dbexporthost, requestid, dbids);
  fprintf(fp, "  echo waiting for jsocdb drms_run commit >> /home/jsoc/exports/tmp/%s.runlog \n",requestid);
  fprintf(fp, "  @ WAITCOUNT = $WAITCOUNT - 1\n");
  fprintf(fp, "  if ($WAITCOUNT <= 0) then\n    set DRMS_ERROR = -1\n    break\n  endif\n");
  fprintf(fp, "  sleep 1\nend \n");
  if (requestorid)
     fprintf(fp, "set Notify=`show_info JSOC_DBHOST=%s -q 'jsoc.export_user[? RequestorID=%d ?]' key=Notify %s` \n", dbexporthost, requestorid, dbids);
  fprintf(fp, "set REQDIR = `show_info JSOC_DBHOST=%s -q -p 'jsoc.export[%s]' %s`\n", dbexporthost, requestid, dbids);
  fprintf(fp, "if ($DRMS_ERROR) then\n");
  fprintf(fp, "  set_keys -C JSOC_DBHOST=%s  ds='jsoc.export[%s]' Status=4 %s\n", dbexporthost, requestid, dbids);
  if (requestorid)
     {
     fprintf(fp, "  mail -n -s 'JSOC export FAILED - %s' $Notify <<!\n", requestid);
     fprintf(fp, "Error status returned from DRMS session.\n");
     fprintf(fp, "See log files at http://jsoc.stanford.edu/$REQDIR\n");
     fprintf(fp, "Also complete log file at /home/jsoc/exports/tmp/%s.runlog\n", requestid);
     fprintf(fp, "!\n");
     }
  fprintf(fp, "else\n");
  if (requestorid)
     {
     fprintf(fp, "  mail -n -s 'JSOC export complete - %s' $Notify <<!\n", requestid);
     fprintf(fp, "JSOC export request %s is complete.\n", requestid);
     fprintf(fp, "Results at http://jsoc.stanford.edu/$REQDIR\n");
     fprintf(fp, "!\n");
     }
  fprintf(fp, "  rm /home/jsoc/exports/tmp/%s.recnum\n", requestid);
  fprintf(fp, "  mv /home/jsoc/exports/tmp/%s.reqlog /home/jsoc/exports/tmp/done \n", requestid);
  /* The log after the call to drms_run gets lost because of the following line - should preserve this
   * somehow (but it can't go in the new inner REQDIR because that is read-only after drms_run returns. */
  fprintf(fp, "  mv /home/jsoc/exports/tmp/%s.runlog /home/jsoc/exports/tmp/done \n", requestid);
  fprintf(fp, "endif\n");
  fclose(fp);
  chmod(qsubscript, 0555);
  }

// Security testing.  Make sure DataSet does not contain attempt to run program
//   example: look out for chars that end dataset spec and give command.
int isbadDataSet(char *in)
  {
  return(0);
  }

// Security testing.  Make sure Processing does not contain attempt to run illegal program
int isbadProcessing(char *process)
  {
  return(0);
  }

#include <time.h>
TIME timenow()
  {
  TIME UNIX_epoch = -220924792.000; /* 1970.01.01_00:00:00_UTC */
  TIME now = (double)time(NULL) + UNIX_epoch;
  return(now);
  }

static void GenErrChkCmd(FILE *fptr)
{
   fprintf(fptr, "set RUNSTAT = $status\nif ($RUNSTAT) goto EXITPLACE\n");
}

/* returns 1 on error, 0 on success */
static int GenHgPatchCmd(FILE *fptr, 
                         const char *hgparms,
                         const char *dbmainhost,
                         const char *dataset,
                         int RecordLimit,
                         const char *requestid,
                         const char *dbids)
{
   int rv = 0;
   char *p = NULL;
   char *hgparams = strdup(hgparms);

   if (hgparams)
   {
      for (p=hgparams; *p; p++)
        if (*p == ',') *p = ' ';
      
      /* Phil says it is okay to remove the hard-coding of his home dir */
      //      fprintf(fptr, "/home/phil/cvs/JSOC/bin/linux_x86_64/hg_patch JSOC_DBHOST=%s %s in='%s' n=%d requestid='%s' log=hg_patch.log %s\n",
      fprintf(fptr, "hg_patch JSOC_DBHOST=%s %s in='%s' n=%d requestid='%s' log=hg_patch.log %s\n",
              dbmainhost, hgparams, dataset, RecordLimit, requestid,  dbids);
      GenErrChkCmd(fptr);
   }
   else
   {
      fprintf(stderr, "XX jsoc_export_manage FAIL - out of memory.\n");
      rv = 1;
   }

   return rv;
}

/* returns 1 on error, 0 on success */
static int GenExpFitsCmd(FILE *fptr, 
                         const char *proto,
                         const char *dbmainhost,
                         const char *requestid,
                         const char *dataset,
                         int RecordLimit,
                         const char *filenamefmt,
                         const char *method,
                         const char *dbids,
                         Processing_t proctype)
{
   char *protocol = strdup(proto);
   int rv = 0;

   if (protocol)
   {
      char *cparms, *p = index(protocol, ',');

      /* The "Protocol" field of jsoc.export has been overloaded. It contains not only the export protocol 
       * (i.e., FITS, as-is, JPEG, MOVIE), but for the FITS protocol, the comma-separated compression 
       * parameters follow. */

      if (p)
      {
         *p = '\0';
         cparms = p+1;
      }
      else
        cparms = "**NONE**";

      if (proctype == kProc_HgPatch)
      {
         /* no "n=XX" */
         fprintf(fptr, "jsoc_export_as_fits JSOC_DBHOST=%s reqid='%s' expversion=%s rsquery='%s' path=$REQDIR ffmt='%s' "
                 "method='%s' protocol='%s' cparms='%s' %s\n",
                 dbmainhost, requestid, PACKLIST_VER, dataset, filenamefmt, method, protos[kProto_FITS], cparms, dbids);
      }
      else
      {
         fprintf(fptr, "jsoc_export_as_fits JSOC_DBHOST=%s reqid='%s' expversion=%s rsquery='%s' n=%d path=$REQDIR ffmt='%s' "
                 "method='%s' protocol='%s' cparms='%s' %s\n",
                 dbmainhost, requestid, PACKLIST_VER, dataset, RecordLimit, filenamefmt, method, protos[kProto_FITS], cparms, dbids);
      }

      GenErrChkCmd(fptr);
   }
   else
   {
      fprintf(stderr, "XX jsoc_export_manage FAIL - out of memory.\n");
      rv = 1;
   }

   return rv;
}

/* returns 1 on error, 0 on success */
static int GenPreProcessCmd(FILE *fptr, 
                            const char *protocol,
                            const char *process,
                            const char *dbmainhost,
                            const char *dataset,
                            int RecordLimit,
                            const char *requestid,
                            const char *method,
                            const char *dbids,
                            Processing_t *proctype,
                            char **dsetout)
{
   int rv = 0;
   char hgds[PATH_MAX];

   if (strncasecmp(process, procs[kProc_HgPatch], strlen(procs[kProc_HgPatch])) == 0)
   {
      *proctype = kProc_HgPatch;

      /* hg_patch compatible with all protocols (if not fits, then assume as-is). */

      /* hg_patch requires additional arguments */
      char *hgparms = NULL;

      if ((hgparms = index(process, ',')) != 0)
      {
         hgparms++;

         if (!(rv = (GenHgPatchCmd(fptr, hgparms, dbmainhost, dataset, RecordLimit, requestid, dbids) != 0)))
         {
            snprintf(hgds, sizeof(hgds), "@%s", kHgPatchLog);
            if (dsetout)
            {
               *dsetout = strdup(hgds);
            }
         }
      }
      else
      {
         /* arguments missing */
         fprintf(stderr,
                 "XX jsoc_export_manage FAIL NOT implemented yet, requestid=%s, process=%s, protocol=%s, method=%s\n",
                 requestid, process, protocol, method);
         rv = 1;
      }
   }
   else if (strncasecmp(process, procs[kProc_SuExport], strlen(procs[kProc_SuExport])) == 0)
   {
      *proctype = kProc_SuExport;
      fprintf(fptr, "jsoc_export_SU_as_is_sock ds='%s' requestid=%s\n", dataset, requestid); 
      GenErrChkCmd(fptr);
      /* rv = 0 */
   }
   else if (strncasecmp(process, procs[kProc_Noop], strlen(procs[kProc_Noop])) == 0)
   {
      *proctype = kProc_Noop;
      /* rv = 0 */
   }
   else if (strncasecmp(process, procs[kProc_NotSpec], strlen(procs[kProc_NotSpec])) == 0)
   {
      *proctype = kProc_NotSpec;
      /* rv = 0 */
   }
   else
   {
      /* Unknown processing */
      fprintf(stderr,
              "XX jsoc_export_manage FAILURE; invalid processing request, requestid=%s, process=%s, protocol=%s, method=%s\n",
              requestid, process, protocol, method);
      rv = 1;
   }

   return rv;
}

static int GenProtoExpCmd(FILE *fptr, 
                          const char *protocol,
                          const char *process,
                          const char *dbmainhost,
                          const char *dataset,
                          int RecordLimit,
                          const char *requestid,
                          const char *method,
                          const char *filenamefmt,
                          const char *dbids,
                          Processing_t proctype)
{
   int rv = 0;

   if (strncasecmp(protocol, protos[kProto_FITS], strlen(protos[kProto_FITS])) == 0)
   {
      rv = (GenExpFitsCmd(fptr, protocol, dbmainhost, requestid, dataset, RecordLimit, filenamefmt, method, dbids, proctype) != 0);
   }
   else if (strncasecmp(protocol, protos[kProto_MPEG], strlen(protos[kProto_MPEG])) == 0)
   {
      /* No "n=XX" */
      /* The arguments are all position-dependent. */
      fprintf(fptr, 
              "jsoc_export_as_movie '%s' '%s' '%s' '%s' '%s' $REQDIR '%s' '%s'\n",
              dataset,
              requestid,
              PACKLIST_VER,
              method,
              protos[kProto_MPEG],
              filenamefmt,
              "cparms is not needed");
   }
   else if (strncasecmp(protocol, protos[kProto_JPEG], strlen(protos[kProto_JPEG])) == 0)
   {
      /* No "n=XX" */
      /* The arguments are all position-dependent. */
      fprintf(fptr, 
              "jsoc_export_as_images '%s' '%s' '%s' '%s' '%s' $REQDIR '%s' '%s'\n",
              dataset,
              requestid,
              PACKLIST_VER,
              method,
              protos[kProto_JPEG],
              filenamefmt,
              "cparms is not needed");
   }
   else if (strncasecmp(protocol, protos[kProto_PNG], strlen(protos[kProto_PNG])) == 0)
   {
      /* No "n=XX" */
      /* The arguments are all position-dependent. */
      fprintf(fptr, 
              "jsoc_export_as_images '%s' '%s' '%s' '%s' '%s' $REQDIR '%s' '%s'\n",
              dataset,
              requestid,
              PACKLIST_VER,
              method,
              protos[kProto_PNG],
              filenamefmt,
              "cparms is not needed");
   }
   else if (strncasecmp(protocol, protos[kProto_AsIs], strlen(protos[kProto_AsIs])) == 0)
   {
      if (proctype == kProc_HgPatch)
      {
         /* No "n=XX" */
         fprintf(fptr, "jsoc_export_as_is JSOC_DBHOST=%s ds='%s' requestid='%s' method='%s' protocol='%s' filenamefmt='%s'\n",
                 dbmainhost, dataset, requestid, method, protos[kProto_AsIs], filenamefmt);
      }
      else
      {
         fprintf(fptr, "jsoc_export_as_is JSOC_DBHOST=%s ds='%s' n=%d requestid='%s' method='%s' protocol='%s' filenamefmt='%s'\n",
                 dbmainhost, dataset, RecordLimit, requestid, method, protos[kProto_AsIs], filenamefmt);
      }

      GenErrChkCmd(fptr);

      /* There is some processing-specific stuff too. */
      if (proctype == kProc_HgPatch)
      {
         fprintf(fptr, "set RECSMADE = `cat hg_patch.log | wc -l` \n");
         GenErrChkCmd(fptr);
         fprintf(fptr, "show_info JSOC_DBHOST=%s n=$RECSMADE -aAit ds='@hg_patch.log' > %s.keywords.txt\n", 
                 dbmainhost, requestid);
         GenErrChkCmd(fptr);
      }
      else if (proctype == kProc_Noop || proctype == kProc_NotSpec)
      {
         fprintf(fptr, "show_info JSOC_DBHOST=%s -ait ds='%s' n=%d > %s.keywords.txt\n",
                 dbmainhost, dataset, RecordLimit, requestid);
         GenErrChkCmd(fptr);
      }
      else
      {
         /* error - can't use this type of processing with as-is */
         fprintf(stderr,
                 "XX jsoc_export_manage FAILURE; invalid protocol/processing type combination, requestid=%s, process=%s, protocol=%s, method=%s\n",
                 requestid, process, protocol, method);
      }
   }
   else
   {
      /* Unknown protocol */
      fprintf(stderr,
              "XX jsoc_export_manage FAILURE; invalid protocol, requestid=%s, process=%s, protocol=%s, method=%s\n",
              requestid, process, protocol, method);
      rv = 1;
   }

   return rv;
}

/* Module main function. */
int DoIt(void)
  {
						/* Get command line arguments */
  char *op;
  char *dataset;
  char *requestid;
  char *process;
  char *requestor;
  long requestorid;
  char *notify;
  char *format;
  char *shipto;
  char *method;
  char *protocol;
  char *filenamefmt;
  char *errorreply;
  char reqdir[DRMS_MAXPATHLEN];
  char command[DRMS_MAXPATHLEN];
  int RecordLimit;
  long long size;
  TIME reqtime;
  TIME esttime;
  TIME exptime;
  TIME now;
  double waittime;
  DRMS_RecordSet_t *exports, *exports_new_orig, *exports_new;
  DRMS_RecordSet_t *requestor_rs;
  DRMS_Record_t *export_rec, *export_log;
  char requestorquery[DRMS_MAXQUERYLEN];
  int status = 0;
  FILE *fp;
  char runscript[DRMS_MAXPATHLEN];
  char *dashp;
  int testmode = 0;

  const char *dbname = NULL;
  const char *dbexporthost = NULL;
  const char *dbmainhost = NULL;
  const char *dbuser = NULL;
  const char *jsocroot = NULL;
  char dbids[128];
  char jsocrootstr[128] = {0};
  int pc = 0;

  int procerr;
  Processing_t proctype;
  char *dset = NULL;

  if (nice_intro ()) return (0);

  testmode = (TESTMODE || cmdparams_isflagset(&cmdparams, kArgTestmode));

  if ((dbmainhost = cmdparams_get_str(&cmdparams, "JSOC_DBMAINHOST", NULL)) == NULL)
  {
     dbmainhost = SERVER;
  }

  if ((dbexporthost = cmdparams_get_str(&cmdparams, "JSOC_DBHOST", NULL)) == NULL)
  {
     dbexporthost = SERVER;
  }

// TEST
// dbmainhost = dbexporthost;

  if ((dbname = cmdparams_get_str(&cmdparams, "JSOC_DBNAME", NULL)) == NULL)
  {
     dbname = DBNAME;
  }

  if ((dbuser = cmdparams_get_str(&cmdparams, "JSOC_DBUSER", NULL)) == NULL)
  {
     dbuser = USER;
  }

  if ((jsocroot = cmdparams_get_str(&cmdparams, "JSOCROOT", NULL)) != NULL)
  {
     snprintf(jsocrootstr, sizeof(jsocrootstr), "JSOCROOT_EXPORT=%s", jsocroot);
  }

  if (dbname)
  {
     pc += snprintf(dbids + pc, sizeof(dbids) - pc, "JSOC_DBNAME=%s ", dbname);
  }

  if (dbuser)
  {
     pc += snprintf(dbids + pc, sizeof(dbids) - pc, "JSOC_DBUSER=%s ", dbuser);
  }

  op = (char *)cmdparams_get_str (&cmdparams, "op", NULL);

  /*  op == process, this is export_manage cmd line, NOT for request being managed */
  if (strcmp(op,"process") == 0) 
    {
    int irec;
    if (testmode)
      exports_new_orig = drms_open_records(drms_env, EXPORT_SERIES_NEW"[][? Status=12 ?]", &status);
    else
      exports_new_orig = drms_open_records(drms_env, EXPORT_SERIES_NEW"[][? Status=2 ?]", &status);
    if (!exports_new_orig)
	DIE("Can not open RecordSet");
    if (exports_new_orig->n < 1)  // No new exports to process.
        {
        drms_close_records(exports_new_orig, DRMS_FREE_RECORD);
        return(0);
        }
    exports_new = drms_clone_records(exports_new_orig, DRMS_PERMANENT, DRMS_SHARE_SEGMENTS, &status);
    if (!exports_new)
	DIE("Can not clone RecordSet");
    drms_close_records(exports_new_orig, DRMS_FREE_RECORD);

    for (irec=0; irec < exports_new->n; irec++) 
      {
      now = timenow();
      export_log = exports_new->records[irec];
      // Get user provided new export request
      status     = drms_getkey_int(export_log, "Status", NULL);
      requestid    = drms_getkey_string(export_log, "RequestID", NULL);
      dataset    = drms_getkey_string(export_log, "DataSet", NULL);
      process    = drms_getkey_string(export_log, "Processing", NULL);
      protocol   = drms_getkey_string(export_log, "Protocol", NULL);
      filenamefmt= drms_getkey_string(export_log, "FilenameFmt", NULL);
      method     = drms_getkey_string(export_log, "Method", NULL);
      format     = drms_getkey_string(export_log, "Format", NULL);
      reqtime    = drms_getkey_time(export_log, "ReqTime", NULL);
      esttime    = drms_getkey_time(export_log, "EstTime", NULL); // Crude guess for now
      size       = drms_getkey_longlong(export_log, "Size", NULL);
      requestorid = drms_getkey_int(export_log, "Requestor", NULL);
      printf("New Request #%d/%d: %s, Status=%d, Processing=%s, DataSet=%s, Protocol=%s, Method=%s\n",
	irec, exports_new->n, requestid, status, process, dataset, protocol, method);
      fflush(stdout);

      // Get user notification email address
      sprintf(requestorquery, "%s[? RequestorID = %ld ?]", EXPORT_USER, requestorid);
      requestor_rs = drms_open_records(drms_env, requestorquery, &status);
      if (!requestor_rs)
        DIE("Cant find requestor info series");
      if (requestor_rs->n > 0)
        {
        DRMS_Record_t *rec = requestor_rs->records[0];
        notify = drms_getkey_string(rec, "Notify", NULL);
        if (*notify == '\0')
           notify = NULL;
        }
      else
        notify = NULL;
      drms_close_records(requestor_rs, DRMS_FREE_RECORD);

      // Create new record in export control series, this one must be DRMS_PERMANENT
      // It will contain the scripts to do the export and set the status to processing
      /* EXPORT_SERIES is jsoc.export. */
      export_rec = drms_create_record(drms_env, EXPORT_SERIES, DRMS_PERMANENT, &status);
      if (!export_rec)
        DIE("Cant create export control record");
  
      drms_setkey_string(export_rec, "RequestID", requestid);
      drms_setkey_string(export_rec, "DataSet", dataset);
      drms_setkey_string(export_rec, "Processing", process);
      drms_setkey_string(export_rec, "Protocol", protocol);
      drms_setkey_string(export_rec, "FilenameFmt", filenamefmt);
      drms_setkey_string(export_rec, "Method", method);
      drms_setkey_string(export_rec, "Format", format);
      drms_setkey_time(export_rec, "ReqTime", now);
      drms_setkey_time(export_rec, "EstTime", now+10); // Crude guess for now
      drms_setkey_longlong(export_rec, "Size", size);
      drms_setkey_int(export_rec, "Requestor", requestorid);
  
      // check  security risk dataset spec or processing request
      if (isbadDataSet(dataset) || isbadProcessing(process))
        { 
        fprintf(stderr," Illegal format detected - security risk!\n"
  		     "RequestID= %s\n"
                       " Processing = %s\n, DataSet=%s\n",
  		     requestid, process, dataset);
        drms_setkey_int(export_rec, "Status", 4);
        drms_close_record(export_rec, DRMS_INSERT_RECORD);
        continue;
        }
  
      drms_record_directory(export_rec, reqdir, 1);

      // Insert qsub command to execute processing script into SU
      make_qsub_call(requestid, reqdir, (notify ? requestorid : 0), dbname, dbuser, dbids, dbexporthost, dbmainhost);
  
      // Insert export processing drms_run script into export record SU
      // The script components must clone the export record with COPY_SEGMENTS in the first usage
      // and with SHARE_SEGMENTS in subsequent modules in the script.  All but the last module
      // in the script may clone as DRMS_TRANSIENT.
      // Remember all modules in this script that deal with the export record must be _sock modules.
      // But modules that need the main database for processing or extracting data should run as
      // direct-connect modules.  They run with the export SU as current directory and must pass all
      // results into that directory.

      // First, prepare initial part of script, same for all processing.
      sprintf(runscript, "%s/%s.drmsrun", reqdir, requestid);
      fp = fopen(runscript, "w");
      fprintf(fp, "#! /bin/csh -f\n");
      fprintf(fp, "set echo\n");
      fprintf(fp, "set histchars\n");
      // force clone with copy segment. 
      fprintf(fp, "set_keys_sock -C JSOC_DBHOST=%s ds='jsoc.export[%s]' Status=1\n", dbexporthost, requestid);
      fprintf(fp, "set RUNSTAT = $status\nif ($RUNSTAT) goto EXITPLACE\n");
      // Get new SU for the export record
      fprintf(fp, "set REQDIR = `show_info_sock JSOC_DBHOST=%s -q -p 'jsoc.export[%s]'`\n", dbexporthost, requestid);
      fprintf(fp, "set RUNSTAT = $status\nif ($RUNSTAT) goto EXITPLACE\n");
      // cd to SU in export record
      fprintf(fp, "cd $REQDIR\n");
      fprintf(fp, "set RUNSTAT = $status\nif ($RUNSTAT) goto EXITPLACE\n");
      // save some diagnostic info
      //fprintf(fp, "printenv > %s.env\n", requestid);
      fprintf(fp, "echo Node = $HOSTNAME\n");
      fprintf(fp, "echo JSOC_DBHOST = %s, Processing DBHOST = %s\n", dbexporthost, dbmainhost);
      fprintf(fp, "echo SUdir = $REQDIR\n");

      // Now generate specific processing related commands

      // extract optional record count limit from process field.
      /* 'process' contains the contents of the 'Processing' column in jsoc.export. Originally, it 
       * used to contain strings like "no_op" or "hg_patch". But now it optionally starts with a 
       * "n=XX" string that is used to limit the number of records returned (in a way completely 
       * analogous to the n=XX parameter to show_info). So the "Processing" field got overloaded.
       * The following block of code checks for the presence of a "n=XX" prefix and strips it off
       * if it is present (setting RecordLimit in the process). */
      if (process && strncmp(process,"n=",2)==0)
        {
        char *now_at = NULL;
        RecordLimit = strtol(process+2, &now_at, 10);
        if (now_at && *now_at == ',')
        process = now_at+1;
fprintf(stderr,"XX Dealing with process=n=xx, RecordLimit=%d, new process=%s\n",RecordLimit,process);
        }
      else
        RecordLimit = 0;

      /* PRE-PROCESSING */

      /* First do the pre-processing of one dataseries into another (if requested). For example, the user may have 
       * requested region extraction. The pre-processing will first make a series of extracted regions 
       * (if it doesn't already exist). The export code will then export from the series of extracted regions, 
       * not from the original full images. */
      procerr = 0;
      proctype = kProc_Unk;
      dset = NULL;

      procerr = GenPreProcessCmd(fp,
                                 protocol,
                                 process, 
                                 dbmainhost, 
                                 dataset, 
                                 RecordLimit, 
                                 requestid, 
                                 method, 
                                 dbids, 
                                 &proctype, 
                                 &dset);

      if (procerr)
      {
         drms_setkey_int(export_log, "Status", 4);
         drms_close_record(export_rec, DRMS_FREE_RECORD);
         fclose(fp);
         fp = NULL;
         if (dataset)
         {
            free(dataset); /* this was malloc'd in this loop */
         }

         if (dset)
         {
            free(dset);
            dset = NULL;
         }
         continue;
      }
      else
      {
         /* If dset was set, then this record-set query should be used in subsequent calls. 
          * As of this writing, dset gets set only if the processing is hg_patch. */
         if (dset)
         {
            if (dataset)
            {
               free(dataset); /* this was malloc'd in this loop */
            }
            dataset = dset;
         }
      }

      /* PROTOCOL-SPECIFIC EXPORT */
      procerr = GenProtoExpCmd(fp, 
                               protocol,
                               process, 
                               dbmainhost, 
                               dataset, 
                               RecordLimit, 
                               requestid, 
                               method, 
                               filenamefmt, 
                               dbids, 
                               proctype);

      if (procerr)
      {
         drms_setkey_int(export_log, "Status", 4);
         drms_close_record(export_rec, DRMS_FREE_RECORD);
         fclose(fp);
         fp = NULL;
         if (dataset)
         {
            free(dataset); /* this was malloc'd in this loop */
         }

         if (dset)
         {
            free(dset);
            dset = NULL;
         }
         continue;
      }
 
      // Finally, add tail of script used for all processing types.
      // convert index.txt list into index.json and index.html packing list files. 
      fprintf(fp, "jsoc_export_make_index\n");
      fprintf(fp, "set RUNSTAT = $status\nif ($RUNSTAT) goto EXITPLACE\n");
      // create tar file if '-tar' suffix on method
      dashp = rindex(method, '-');
      if (dashp && strcmp(dashp, "-tar") == 0)
        {
        fprintf(fp, "cp %s.* index.* ..\n", requestid);
        fprintf(fp, "tar  chf ../%s.tar --remove-files ./\n", requestid);
        fprintf(fp, "set RUNSTAT = $status\nif ($RUNSTAT) goto EXITPLACE\n");
        fprintf(fp, "mv ../%s.* ../index.* .\n", requestid);
        fprintf(fp, "set RUNSTAT = $status\nif ($RUNSTAT) goto EXITPLACE\n");
        }

      // DONE, Standard exit here only if no errors above
      // set status=done and mark this version of the export record permanent
      fprintf(fp, "set_keys_sock JSOC_DBHOST=%s ds='jsoc.export[%s]' Status=0\n", dbexporthost, requestid);
                  // copy the drms_run log file
      fprintf(fp, "cp /home/jsoc/exports/tmp/%s.runlog ./%s.runlog \n", requestid, requestid);

      // DONE, Standard exit here if errors, falls through to here if OK
      fprintf(fp, "EXITPLACE:\n");
      fprintf(fp, "if ($RUNSTAT) then\n");
      fprintf(fp, "  echo XXXXXXXXXXXXXXXXXXXXXXX ERROR EXIT XXXXXXXXXXXXXXXXXXXXXXXXXXX\nprintenv\n");
      fprintf(fp, "endif\n");
                  // make drms_run completion lock file (always do this)
      fprintf(fp, "show_info_sock JSOC_DBHOST=%s -q -r 'jsoc.export[%s]' > /home/jsoc/exports/tmp/%s.recnum \n", dbexporthost, requestid, requestid);
      fprintf(fp, "set LOCKSTAT = $status\n");
      fprintf(fp, "if ($LOCKSTAT) then\n");
                  // make drms_run completion lock file (always do this) - drms_server died, so don't use sock version here
      fprintf(fp, "  show_info JSOC_DBHOST=%s -q -r 'jsoc.export[%s]' > /home/jsoc/exports/tmp/%s.recnum \n", dbexporthost, requestid, requestid);
      fprintf(fp, "endif\n");
      fprintf(fp, "exit $RUNSTAT\n");
      fclose(fp);
      chmod(runscript, 0555);
  
      // close the current (first) version of the record in jsoc.export
// fprintf(stderr,"export_manage closing new record for %s\n",requestid);
      drms_setkey_int(export_rec, "Status", 1);
      drms_close_record(export_rec, DRMS_INSERT_RECORD);
// fprintf(stderr,"export_manage closed new record for %s, starting qsub\n",requestid);
  
      // SU now contains both qsub script and drms_run script, ready to execute and lock the record.
      //sprintf(command,"qsub -q x.q,o.q,j.q -v %s "
      sprintf(command,"qsub -q j.q -v %s "
  	" -o /home/jsoc/exports/tmp/%s.runlog "
  	" -e /home/jsoc/exports/tmp/%s.runlog "
  	"  %s/%s.qsub ",
        jsocrootstr, requestid, requestid, reqdir, requestid);
  /*
  	"  >>& /home/jsoc/exports/tmp/%s.runlog",
  */
// fprintf(stderr,"export_manage for %s, qsub=%s\n",requestid, command);
      if (system(command))
        DIE("Submission of qsub command failed");
  
      // Mark the record in jsoc.export_new as accepted for processing.
      drms_setkey_int(export_log, "Status", 1);
      printf("Request %s submitted\n",requestid);
      } // end looping on new export requests (looping over records in jsoc.export_new)

    drms_close_records(exports_new, DRMS_INSERT_RECORD);
    return(0);
    } // End process new requests.
  else if (strcmp(op, "SOMETHINGELSE") == 0)
    {
    }
  else 
    DIE("Operation not allowed");
  return(1);
  }

